# Copyright (C) 2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
#                         Nikos Kouremenos <kourem AT gmail.com>
#                         Travis Shirk <travis AT pobox.com>
# Copyright (C) 2007 Lukas Petrovicky <lukas AT petrovicky.net>
#                    Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
#                         Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim. If not, see <http://www.gnu.org/licenses/>.

from typing import ClassVar  # pylint: disable=unused-import
from typing import Type  # pylint: disable=unused-import
from typing import Optional  # pylint: disable=unused-import

import os
import time
import logging

from gi.repository import Gtk
from gi.repository import Gio
from gi.repository import Pango
from gi.repository import GLib
from gi.repository import Gdk

from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common import helpers
from gajim.common import ged
from gajim.common import i18n
from gajim.common.i18n import _
from gajim.common.helpers import AdditionalDataDict
from gajim.common.helpers import open_uri
from gajim.common.helpers import geo_provider_from_location
from gajim.common.helpers import event_filter
from gajim.common.helpers import open_file
from gajim.common.contacts import GC_Contact
from gajim.common.const import AvatarSize
from gajim.common.const import KindConstant
from gajim.common.const import Chatstate
from gajim.common.const import PEPEventType
from gajim.common.const import JingleState

from gajim import gtkgui_helpers
from gajim import gui_menu_builder
from gajim import dialogs

from gajim.gui.gstreamer import create_gtk_widget
from gajim.gui.dialogs import DialogButton
from gajim.gui.dialogs import ConfirmationDialog
from gajim.gui.add_contact import AddNewContactWindow
from gajim.gui.util import get_cursor
from gajim.gui.util import format_mood
from gajim.gui.util import format_activity
from gajim.gui.util import format_tune
from gajim.gui.util import format_location
from gajim.gui.util import get_activity_icon_name
from gajim.gui.util import make_href_markup
from gajim.gui.const import ControlType

from gajim.command_system.implementation.hosts import ChatCommands
from gajim.command_system.framework import CommandHost  # pylint: disable=unused-import
from gajim.chat_control_base import ChatControlBase

log = logging.getLogger('gajim.chat_control')


class JingleObject:
    __slots__ = ('sid', 'state', 'available', 'update')

    def __init__(self, state, update):
        self.sid = None
        self.state = state
        self.available = False
        self.update = update


################################################################################
class ChatControl(ChatControlBase):
    """
    A control for standard 1-1 chat
    """
    _type = ControlType.CHAT
    old_msg_kind = None # last kind of the printed message

    # Set a command host to bound to. Every command given through a chat will be
    # processed with this command host.
    COMMAND_HOST = ChatCommands  # type: ClassVar[Type[CommandHost]]

    def __init__(self, parent_win, contact, acct, session, resource=None):
        ChatControlBase.__init__(self,
                                 parent_win,
                                 'chat_control',
                                 contact,
                                 acct,
                                 resource)

        self.last_recv_message_id = None
        self.last_recv_message_marks = None
        self.last_message_timestamp = None

        self.toggle_emoticons()

        if not app.settings.get('hide_chat_banner'):
            self.xml.banner_eventbox.set_no_show_all(False)

        self.xml.sendfile_button.set_action_name(
            'win.send-file-%s' % self.control_id)

        # Menu for the HeaderBar
        self.control_menu = gui_menu_builder.get_singlechat_menu(
            self.control_id, self.account, self.contact.jid, self._type)

        # Settings menu
        self.xml.settings_menu.set_menu_model(self.control_menu)

        self.jingle = {
            'audio': JingleObject(
                JingleState.NULL,
                self.update_audio),
            'video': JingleObject(
                JingleState.NULL,
                self.update_video),
        }
        self._video_widget_other = None
        self._video_widget_self = None

        self.update_toolbar()
        self.update_all_pep_types()
        self._update_avatar()

        # Hook up signals
        widget = self.xml.location_eventbox
        id_ = widget.connect('button-release-event',
                             self.on_location_eventbox_button_release_event)
        self.handlers[id_] = widget
        id_ = widget.connect('enter-notify-event',
                             self.on_location_eventbox_enter_notify_event)
        self.handlers[id_] = widget
        id_ = widget.connect('leave-notify-event',
                             self.on_location_eventbox_leave_notify_event)
        self.handlers[id_] = widget

        for key in ('1', '2', '3', '4', '5', '6', '7', '8', '9', '*', '0', '#'):
            widget = self.xml.get_object(key + '_button')
            id_ = widget.connect('pressed', self.on_num_button_pressed, key)
            self.handlers[id_] = widget
            id_ = widget.connect('released', self.on_num_button_released)
            self.handlers[id_] = widget

        widget = self.xml.mic_hscale
        id_ = widget.connect('value_changed', self.on_mic_hscale_value_changed)
        self.handlers[id_] = widget

        widget = self.xml.sound_hscale
        id_ = widget.connect('value_changed',
                             self.on_sound_hscale_value_changed)
        self.handlers[id_] = widget

        self.info_bar = Gtk.InfoBar()
        content_area = self.info_bar.get_content_area()
        self.info_bar_label = Gtk.Label()
        self.info_bar_label.set_use_markup(True)
        self.info_bar_label.set_halign(Gtk.Align.START)
        self.info_bar_label.set_valign(Gtk.Align.START)
        self.info_bar_label.set_ellipsize(Pango.EllipsizeMode.END)
        content_area.add(self.info_bar_label)
        self.info_bar.set_no_show_all(True)

        self.xml.vbox2.pack_start(self.info_bar, False, True, 5)
        self.xml.vbox2.reorder_child(self.info_bar, 1)

        # List of waiting infobar messages
        self.info_bar_queue = []

        self.subscribe_events()

        if not session:
            # Don't use previous session if we want to a specific resource
            # and it's not the same
            if not resource:
                resource = contact.resource
            session = app.connections[self.account].find_controlless_session(
                self.contact.jid, resource)

        self.setup_seclabel()
        if session:
            session.control = self
            self.session = session

        self.add_actions()
        self.update_ui()
        self.set_lock_image()

        self.xml.encryption_menu.set_menu_model(
            gui_menu_builder.get_encryption_menu(
                self.control_id, self._type, self.account == 'Local'))
        self.set_encryption_menu_icon()
        # restore previous conversation
        self.restore_conversation()
        self.msg_textview.grab_focus()

        # pylint: disable=line-too-long
        self.register_events([
            ('nickname-received', ged.GUI1, self._on_nickname_received),
            ('mood-received', ged.GUI1, self._on_mood_received),
            ('activity-received', ged.GUI1, self._on_activity_received),
            ('tune-received', ged.GUI1, self._on_tune_received),
            ('location-received', ged.GUI1, self._on_location_received),
            ('update-client-info', ged.GUI1, self._on_update_client_info),
            ('chatstate-received', ged.GUI1, self._on_chatstate_received),
            ('caps-update', ged.GUI1, self._on_caps_update),
            ('message-sent', ged.OUT_POSTCORE, self._on_message_sent),
            ('mam-decrypted-message-received', ged.GUI1, self._on_mam_decrypted_message_received),
            ('decrypted-message-received', ged.GUI1, self._on_decrypted_message_received),
            ('receipt-received', ged.GUI1, self._receipt_received),
            ('displayed-received', ged.GUI1, self._displayed_received),
            ('message-error', ged.GUI1, self._on_message_error),
            ('zeroconf-error', ged.GUI1, self._on_zeroconf_error),
        ])

        if self._type.is_chat:
            # Don’t connect this when PrivateChatControl is used
            self.register_event('update-roster-avatar', ged.GUI1, self._on_update_roster_avatar)
        # pylint: enable=line-too-long

        # PluginSystem: adding GUI extension point for this ChatControl
        # instance object
        app.plugin_manager.gui_extension_point('chat_control', self)
        self.update_actions()

    @property
    def jid(self):
        return self.contact.jid

    def add_actions(self):
        super().add_actions()
        actions = [
            ('invite-contacts-', self._on_invite_contacts),
            ('add-to-roster-', self._on_add_to_roster),
            ('block-contact-', self._on_block_contact),
            ('information-', self._on_information),
            ('start-call-', self._on_start_call),
        ]

        for action in actions:
            action_name, func = action
            act = Gio.SimpleAction.new(action_name + self.control_id, None)
            act.connect('activate', func)
            self.parent_win.window.add_action(act)

        chatstate = self.contact.settings.get('send_chatstate')

        act = Gio.SimpleAction.new_stateful(
            'send-chatstate-' + self.control_id,
            GLib.VariantType.new("s"),
            GLib.Variant("s", chatstate))
        act.connect('change-state', self._on_send_chatstate)
        self.parent_win.window.add_action(act)

        marker = self.contact.settings.get('send_marker')

        act = Gio.SimpleAction.new_stateful(
            f'send-marker-{self.control_id}',
            None,
            GLib.Variant.new_boolean(marker))
        act.connect('change-state', self._on_send_marker)
        self.parent_win.window.add_action(act)

    def update_actions(self):
        win = self.parent_win.window
        online = app.account_is_connected(self.account)
        con = app.connections[self.account]

        # Add to roster
        if not isinstance(self.contact, GC_Contact) \
        and _('Not in contact list') in self.contact.groups and \
        app.connections[self.account].roster_supported and online:
            win.lookup_action(
                'add-to-roster-' + self.control_id).set_enabled(True)
        else:
            win.lookup_action(
                'add-to-roster-' + self.control_id).set_enabled(False)

        # Block contact
        win.lookup_action(
            'block-contact-' + self.control_id).set_enabled(
                online and con.get_module('Blocking').supported)

        # Jingle AV detection
        if (self.contact.supports(Namespace.JINGLE_ICE_UDP) and
                app.is_installed('FARSTREAM') and self.contact.resource):
            self.jingle['audio'].available = self.contact.supports(
                Namespace.JINGLE_RTP_AUDIO)
            self.jingle['video'].available = self.contact.supports(
                Namespace.JINGLE_RTP_VIDEO)
        else:
            if (self.jingle['audio'].available or
                    self.jingle['video'].available):
                self.stop_jingle()
            self.jingle['audio'].available = False
            self.jingle['video'].available = False

        win.lookup_action(f'start-call-{self.control_id}').set_enabled(
            online and (self.jingle['audio'].available or
                        self.jingle['video'].available))

        # Send message
        has_text = self.msg_textview.has_text()
        win.lookup_action(
            f'send-message-{self.control_id}').set_enabled(online and has_text)

        # Send file (HTTP File Upload)
        httpupload = win.lookup_action(
            'send-file-httpupload-' + self.control_id)
        httpupload.set_enabled(
            online and con.get_module('HTTPUpload').available)

        # Send file (Jingle)
        jingle_support = self.contact.supports(Namespace.JINGLE_FILE_TRANSFER_5)
        jingle_conditions = jingle_support and self.contact.show != 'offline'
        jingle = win.lookup_action('send-file-jingle-' + self.control_id)
        jingle.set_enabled(online and jingle_conditions)

        # Send file
        win.lookup_action(
            'send-file-' + self.control_id).set_enabled(
                jingle.get_enabled() or httpupload.get_enabled())

        # Set File Transfer Button tooltip
        if online and (httpupload.get_enabled() or jingle.get_enabled()):
            tooltip_text = _('Send File…')
        else:
            tooltip_text = _('No File Transfer available')
        self.xml.sendfile_button.set_tooltip_text(tooltip_text)

        # Chat markers
        state = GLib.Variant.new_boolean(
            self.contact.settings.get('send_marker'))
        win.lookup_action(
            f'send-marker-{self.control_id}').change_state(state)

        # Convert to GC
        if app.settings.get_account_setting(self.account, 'is_zeroconf'):
            win.lookup_action(
                'invite-contacts-' + self.control_id).set_enabled(False)
        else:
            if self.contact.supports(Namespace.MUC) and online:
                win.lookup_action(
                    'invite-contacts-' + self.control_id).set_enabled(True)
            else:
                win.lookup_action(
                    'invite-contacts-' + self.control_id).set_enabled(False)

        # Information
        win.lookup_action(
            'information-' + self.control_id).set_enabled(online)

    def remove_actions(self):
        super().remove_actions()
        actions = [
            'invite-contacts-',
            'add-to-roster-',
            'block-contact-',
            'information-',
            'start-call-',
            'send-chatstate-',
            'send-marker-',
        ]
        for action in actions:
            self.parent_win.window.remove_action(f'{action}{self.control_id}')

    def focus(self):
        self.msg_textview.grab_focus()

    def delegate_action(self, action):
        res = super().delegate_action(action)
        if res == Gdk.EVENT_STOP:
            return res

        if action == 'show-contact-info':
            self.parent_win.window.lookup_action(
                'information-%s' % self.control_id).activate()
            return Gdk.EVENT_STOP

        if action == 'send-file':
            if app.interface.msg_win_mgr.mode == \
            app.interface.msg_win_mgr.ONE_MSG_WINDOW_ALWAYS_WITH_ROSTER:
                app.interface.roster.tree.grab_focus()
                return Gdk.EVENT_PROPAGATE

            self.parent_win.window.lookup_action(
                'send-file-%s' % self.control_id).activate()
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_add_to_roster(self, _action, _param):
        AddNewContactWindow(self.account, self.contact.jid)

    def _on_block_contact(self, _action, _param):
        def _block_contact(report=None):
            con = app.connections[self.account]
            con.get_module('Blocking').block([self.contact.jid], report=report)

            self.parent_win.remove_tab(self, None, force=True)
            if _('Not in contact list') in self.contact.get_shown_groups():
                app.interface.roster.remove_contact(
                    self.contact.jid, self.account, force=True, backend=True)
                return
            app.interface.roster.draw_contact(self.contact.jid, self.account)

        ConfirmationDialog(
            _('Block Contact'),
            _('Really block this contact?'),
            _('You will appear offline for this contact and you will '
              'not receive further messages.'),
            [DialogButton.make('Cancel'),
             DialogButton.make('OK',
                               text=_('_Report Spam'),
                               callback=_block_contact,
                               kwargs={'report': 'spam'}),
             DialogButton.make('Remove',
                               text=_('_Block'),
                               callback=_block_contact)],
            modal=False).show()

    def _on_information(self, _action, _param):
        app.interface.roster.on_info(None, self.contact, self.account)

    def _on_invite_contacts(self, _action, _param):
        """
        User wants to invite some friends to chat
        """
        dialogs.TransformChatToMUC(self.account, [self.contact.jid])

    def _on_send_chatstate(self, action, param):
        action.set_state(param)
        self.contact.settings.set('send_chatstate', param.get_string())

    def _on_send_marker(self, action, param):
        action.set_state(param)
        self.contact.settings.set('send_marker', param.get_boolean())

    def subscribe_events(self):
        """
        Register listeners to the events class
        """
        app.events.event_added_subscribe(self.on_event_added)
        app.events.event_removed_subscribe(self.on_event_removed)

    def unsubscribe_events(self):
        """
        Unregister listeners to the events class
        """
        app.events.event_added_unsubscribe(self.on_event_added)
        app.events.event_removed_unsubscribe(self.on_event_removed)

    def _update_toolbar(self):
        # Formatting
        # TODO: find out what encryption allows for xhtml and which not
        if self.contact.supports(Namespace.XHTML_IM):
            self.xml.formattings_button.set_sensitive(True)
            self.xml.formattings_button.set_tooltip_text(_(
                'Show a list of formattings'))
        else:
            self.xml.formattings_button.set_sensitive(False)
            self.xml.formattings_button.set_tooltip_text(
                _('This contact does not support HTML'))

    def update_all_pep_types(self):
        self._update_pep(PEPEventType.LOCATION)
        self._update_pep(PEPEventType.MOOD)
        self._update_pep(PEPEventType.ACTIVITY)
        self._update_pep(PEPEventType.TUNE)

    def _update_pep(self, type_):
        image = self._get_pep_widget(type_)
        data = self.contact.pep.get(type_)
        if data is None:
            image.hide()
            return

        if type_ == PEPEventType.MOOD:
            icon = 'mood-%s' % data.mood
            formated_text = format_mood(*data)
        elif type_ == PEPEventType.ACTIVITY:
            icon = get_activity_icon_name(data.activity, data.subactivity)
            formated_text = format_activity(*data)
        elif type_ == PEPEventType.TUNE:
            icon = 'audio-x-generic'
            formated_text = format_tune(*data)
        elif type_ == PEPEventType.LOCATION:
            icon = 'applications-internet'
            formated_text = format_location(data)

        image.set_from_icon_name(icon, Gtk.IconSize.MENU)
        image.set_tooltip_markup(formated_text)
        image.show()

    def _get_pep_widget(self, type_):
        if type_ == PEPEventType.MOOD:
            return self.xml.mood_image
        if type_ == PEPEventType.ACTIVITY:
            return self.xml.activity_image
        if type_ == PEPEventType.TUNE:
            return self.xml.tune_image
        if type_ == PEPEventType.LOCATION:
            return self.xml.location_image
        return None

    @event_filter(['account', 'jid'])
    def _on_mood_received(self, _event):
        self._update_pep(PEPEventType.MOOD)

    @event_filter(['account', 'jid'])
    def _on_activity_received(self, _event):
        self._update_pep(PEPEventType.ACTIVITY)

    @event_filter(['account', 'jid'])
    def _on_tune_received(self, _event):
        self._update_pep(PEPEventType.TUNE)

    @event_filter(['account', 'jid'])
    def _on_location_received(self, _event):
        self._update_pep(PEPEventType.LOCATION)

    @event_filter(['account', 'jid'])
    def _on_nickname_received(self, _event):
        self.update_ui()
        self.parent_win.redraw_tab(self)
        self.parent_win.show_title()

    @event_filter(['account', 'jid'])
    def _on_update_client_info(self, event):
        contact = app.contacts.get_contact(
            self.account, event.jid, event.resource)
        if contact is None:
            return
        self.xml.phone_image.set_visible(contact.uses_phone)

    @event_filter(['account'])
    def _on_chatstate_received(self, event):
        if self._type.is_privatechat:
            if event.contact != self.gc_contact:
                return
        else:
            if event.contact.jid != self.contact.jid:
                return

        self.draw_banner_text()

        # update chatstate in tab for this chat
        if event.contact.is_gc_contact:
            chatstate = event.contact.chatstate
        else:
            chatstate = app.contacts.get_combined_chatstate(
                self.account, self.contact.jid)
        self.parent_win.redraw_tab(self, chatstate)

    @event_filter(['account'])
    def _on_caps_update(self, event):
        if self._type.is_chat and event.jid != self.contact.jid:
            return
        if self._type.is_privatechat and event.fjid != self.contact.jid:
            return
        self.update_ui()

    @event_filter(['account'])
    def _on_mam_decrypted_message_received(self, event):
        if event.properties.type.is_groupchat:
            return

        if event.properties.is_muc_pm:
            if not event.properties.jid == self.contact.get_full_jid():
                return
        else:
            if not event.properties.jid.bare_match(self.contact.jid):
                return

        kind = '' # incoming
        if event.kind == KindConstant.CHAT_MSG_SENT:
            kind = 'outgoing'

        self.add_message(event.msgtxt,
                         kind,
                         tim=event.properties.mam.timestamp,
                         correct_id=event.correct_id,
                         message_id=event.properties.id,
                         additional_data=event.additional_data)

    @event_filter(['account'])
    def _on_decrypted_message_received(self, event):
        if not event.msgtxt:
            return True

        if event.session.control != self:
            return

        typ = ''
        if event.properties.is_sent_carbon:
            typ = 'out'

        self.add_message(event.msgtxt,
                         typ,
                         tim=event.properties.timestamp,
                         subject=event.properties.subject,
                         displaymarking=event.displaymarking,
                         msg_log_id=event.msg_log_id,
                         message_id=event.properties.id,
                         correct_id=event.correct_id,
                         additional_data=event.additional_data)
        if event.msg_log_id:
            pw = self.parent_win
            end = self.conv_textview.autoscroll
            if not pw or (pw.get_active_control() and self \
            == pw.get_active_control() and pw.is_active() and end):
                app.storage.archive.set_read_messages([event.msg_log_id])

    @event_filter(['account', 'jid'])
    def _on_message_error(self, event):
        self.conv_textview.show_error(event.message_id, event.error)

    @event_filter(['account', 'jid'])
    def _on_message_sent(self, event):
        if not event.message:
            return

        self.last_sent_msg = event.message_id
        message_id = event.message_id

        if event.label:
            displaymarking = event.label.displaymarking
        else:
            displaymarking = None
        if self.correcting:
            self.correcting = False
            gtkgui_helpers.remove_css_class(
                self.msg_textview, 'gajim-msg-correcting')

        self.add_message(event.message,
                         self.contact.jid,
                         tim=event.timestamp,
                         displaymarking=displaymarking,
                         message_id=message_id,
                         correct_id=event.correct_id,
                         additional_data=event.additional_data)

    @event_filter(['account', 'jid'])
    def _receipt_received(self, event):
        self.conv_textview.show_receipt(event.receipt_id)

    @event_filter(['account', 'jid'])
    def _displayed_received(self, event):
        self.conv_textview.show_displayed(event.marker_id)

    @event_filter(['account', 'jid'])
    def _on_zeroconf_error(self, event):
        self.add_status_message(event.message)

    @event_filter(['account', 'jid'])
    def _on_update_roster_avatar(self, obj):
        self._update_avatar()

    @event_filter(['account'])
    def _nec_ping(self, event):
        if self.contact != event.contact:
            return
        if event.name == 'ping-sent':
            self.add_info_message(_('Ping?'))
        elif event.name == 'ping-reply':
            self.add_info_message(
                _('Pong! (%s seconds)') % event.seconds)
        elif event.name == 'ping-error':
            self.add_info_message(event.error)

    def change_resource(self, resource):
        old_full_jid = self.get_full_jid()
        self.resource = resource
        new_full_jid = self.get_full_jid()
        # update app.last_message_time
        if old_full_jid in app.last_message_time[self.account]:
            app.last_message_time[self.account][new_full_jid] = \
                    app.last_message_time[self.account][old_full_jid]
        # update events
        app.events.change_jid(self.account, old_full_jid, new_full_jid)
        # update MessageWindow._controls
        self.parent_win.change_jid(self.account, old_full_jid, new_full_jid)

    # Jingle AV
    def _on_start_call(self, *args):
        audio_state = self.jingle['audio'].state
        video_state = self.jingle['video'].state
        if audio_state == JingleState.NULL and video_state == JingleState.NULL:
            self.xml.av_box.set_no_show_all(False)
            self.xml.av_box.show_all()
            self.xml.jingle_audio_state.hide()
            self.xml.av_start_box.show()
            self.xml.av_start_mic_cam_button.set_sensitive(
                self.jingle['video'].available)
            self.xml.av_cam_button.set_sensitive(False)

    def _on_call_with_mic(self, _button):
        self._on_jingle_button_toggled(['audio'])
        self.xml.av_start_box.hide()

    def _on_call_with_mic_and_cam(self, _button):
        self._on_jingle_button_toggled(['audio', 'video'])
        self.xml.av_start_box.hide()

    def _on_video(self, *args):
        self._on_jingle_button_toggled(['video'])

    def update_audio(self):
        self.update_actions()

        audio_state = self.jingle['audio'].state
        video_state = self.jingle['video'].state
        if self.jingle['video'].available:
            self.xml.av_cam_button.set_sensitive(
                video_state not in (
                    JingleState.CONNECTING,
                    JingleState.CONNECTED))

        if audio_state == JingleState.NULL:
            self.xml.audio_buttons_box.set_sensitive(False)
            self.xml.jingle_audio_state.set_no_show_all(True)
            self.xml.jingle_audio_state.hide()
            self.xml.jingle_connection_state.set_text('')
            self.xml.jingle_connection_spinner.stop()
            self.xml.jingle_connection_spinner.hide()

            if video_state == JingleState.NULL:
                self.xml.av_box.set_no_show_all(True)
                self.xml.av_box.hide()
        else:
            self.xml.jingle_connection_spinner.show()
            self.xml.jingle_connection_spinner.start()

        if audio_state == JingleState.CONNECTING:
            self.xml.av_box.set_no_show_all(False)
            self.xml.av_box.show_all()
            self.xml.jingle_connection_state.set_text(
                _('Calling…'))
            self.xml.av_cam_button.set_sensitive(False)

        elif audio_state == JingleState.CONNECTION_RECEIVED:
            self.xml.jingle_connection_state.set_text(
                _('Incoming Call'))

        elif audio_state == JingleState.CONNECTED:
            self.xml.jingle_audio_state.set_no_show_all(False)
            self.xml.jingle_audio_state.show()
            self.xml.jingle_connection_state.set_text('')
            self.xml.jingle_connection_spinner.stop()
            self.xml.jingle_connection_spinner.hide()
            if self.jingle['video'].available:
                self.xml.av_cam_button.set_sensitive(True)

            input_vol = app.settings.get('audio_input_volume')
            output_vol = app.settings.get('audio_output_volume')
            self.xml.mic_hscale.set_value(max(min(input_vol, 100), 0))
            self.xml.sound_hscale.set_value(max(min(output_vol, 100), 0))
            self.xml.audio_buttons_box.set_sensitive(True)

        elif audio_state == JingleState.ERROR:
            self.xml.jingle_audio_state.hide()
            self.xml.jingle_connection_state.set_text(
                _('Connection Error'))
            self.xml.jingle_connection_spinner.stop()
            self.xml.jingle_connection_spinner.hide()

        if not self.jingle['audio'].sid:
            self.xml.audio_buttons_box.set_sensitive(False)

    def update_video(self):
        self.update_actions()

        audio_state = self.jingle['audio'].state
        video_state = self.jingle['video'].state

        if video_state == JingleState.NULL:
            self.xml.video_box.set_no_show_all(True)
            self.xml.video_box.hide()
            self.xml.outgoing_viewport.set_no_show_all(True)
            self.xml.outgoing_viewport.hide()
            if self._video_widget_other:
                self._video_widget_other.destroy()
            if self._video_widget_self:
                self._video_widget_self.destroy()

            if audio_state != JingleState.CONNECTED:
                self.xml.jingle_connection_state.set_text('')
            self.xml.jingle_connection_spinner.stop()
            self.xml.jingle_connection_spinner.hide()
            self.xml.av_cam_button.set_sensitive(True)
            self.xml.av_cam_button.set_tooltip_text(_('Turn Camera on'))
            self.xml.av_cam_image.set_from_icon_name(
                'feather-camera-symbolic', Gtk.IconSize.BUTTON)
            if audio_state == JingleState.NULL:
                self.xml.av_box.set_no_show_all(True)
                self.xml.av_box.hide()
        else:
            self.xml.jingle_connection_spinner.show()
            self.xml.jingle_connection_spinner.start()

        if video_state == JingleState.CONNECTING:
            self.xml.jingle_connection_state.set_text(_('Calling (Video)…'))
            self.xml.av_box.set_no_show_all(False)
            self.xml.av_box.show_all()
            self.xml.av_cam_button.set_sensitive(False)
            self.xml.av_cam_button.set_tooltip_text(_('Turn Camera off'))
            self.xml.av_cam_image.set_from_icon_name(
                'feather-camera-off-symbolic', Gtk.IconSize.BUTTON)

        elif video_state == JingleState.CONNECTION_RECEIVED:
            self.xml.jingle_connection_state.set_text(
                _('Incoming Call (Video)'))
            self.xml.av_cam_button.set_sensitive(False)
            self.xml.av_cam_button.set_tooltip_text(_('Turn Camera off'))
            self.xml.av_cam_image.set_from_icon_name(
                'feather-camera-off-symbolic', Gtk.IconSize.BUTTON)

        elif video_state == JingleState.CONNECTED:
            self.xml.video_box.set_no_show_all(False)
            self.xml.video_box.show_all()
            if app.settings.get('video_see_self'):
                self.xml.outgoing_viewport.set_no_show_all(False)
                self.xml.outgoing_viewport.show()
            else:
                self.xml.outgoing_viewport.set_no_show_all(True)
                self.xml.outgoing_viewport.hide()

            sink_other, self._video_widget_other, _name = create_gtk_widget()
            sink_self, self._video_widget_self, _name = create_gtk_widget()
            self.xml.incoming_viewport.add(self._video_widget_other)
            self.xml.outgoing_viewport.add(self._video_widget_self)

            con = app.connections[self.account]
            session = con.get_module('Jingle').get_jingle_session(
                self.contact.get_full_jid(), self.jingle['video'].sid)
            content = session.get_content('video')
            content.do_setup(sink_self, sink_other)

            self.xml.jingle_connection_state.set_text('')
            self.xml.jingle_connection_spinner.stop()
            self.xml.jingle_connection_spinner.hide()

            self.xml.av_cam_button.set_sensitive(True)
            self.xml.av_cam_button.set_tooltip_text(_('Turn Camera off'))
            self.xml.av_cam_image.set_from_icon_name(
                'feather-camera-off-symbolic', Gtk.IconSize.BUTTON)

        elif video_state == JingleState.ERROR:
            self.xml.jingle_connection_state.set_text(
                _('Connection Error'))
            self.xml.jingle_connection_spinner.stop()
            self.xml.jingle_connection_spinner.hide()

    def set_jingle_state(self, jingle_type: str, state: str, sid: str = None,
                         reason: str = None) -> None:
        jingle = self.jingle[jingle_type]
        if state in (
                JingleState.CONNECTING,
                JingleState.CONNECTED,
                JingleState.NULL,
                JingleState.ERROR) and reason:
            log.info('%s state: %s, reason: %s', jingle_type, state, reason)

        if state in (jingle.state, JingleState.ERROR):
            return

        if (state == JingleState.NULL and jingle.sid not in (None, sid)):
            return

        new_sid = None
        if state == JingleState.NULL:
            new_sid = None
        if state in (
                JingleState.CONNECTION_RECEIVED,
                JingleState.CONNECTING,
                JingleState.CONNECTED):
            new_sid = sid

        jingle.state = state
        jingle.sid = new_sid
        jingle.update()

    def stop_jingle(self, sid=None, reason=None):
        audio_sid = self.jingle['audio'].sid
        video_sid = self.jingle['video'].sid
        if audio_sid and sid in (audio_sid, None):
            self.close_jingle_content('audio')
        if video_sid and sid in (video_sid, None):
            self.close_jingle_content('video')

    def close_jingle_content(self, jingle_type: str,
                             shutdown: Optional[bool] = False) -> None:
        jingle = self.jingle[jingle_type]
        if not jingle.sid:
            return

        con = app.connections[self.account]
        session = con.get_module('Jingle').get_jingle_session(
            self.contact.get_full_jid(), jingle.sid)
        if session:
            content = session.get_content(jingle_type)
            if content:
                session.remove_content(content.creator, content.name)

        if not shutdown:
            jingle.sid = None
            jingle.state = JingleState.NULL
            jingle.update()

    def _on_end_call_clicked(self, _widget):
        self.close_jingle_content('audio')
        self.close_jingle_content('video')
        self.xml.jingle_audio_state.set_no_show_all(True)
        self.xml.jingle_audio_state.hide()
        self.xml.av_box.set_no_show_all(True)
        self.xml.av_box.hide()

    def _on_jingle_button_toggled(self, jingle_types):
        con = app.connections[self.account]

        if all(item in jingle_types for item in ['audio', 'video']):
            # Both 'audio' and 'video' in jingle_types
            sid = con.get_module('Jingle').start_audio_video(
                self.contact.get_full_jid())
            self.set_jingle_state('audio', JingleState.CONNECTING, sid)
            self.set_jingle_state('video', JingleState.CONNECTING, sid)
            return

        if 'audio' in jingle_types:
            if self.jingle['audio'].state != JingleState.NULL:
                self.close_jingle_content('audio')
            else:
                sid = con.get_module('Jingle').start_audio(
                    self.contact.get_full_jid())
                self.set_jingle_state('audio', JingleState.CONNECTING, sid)

        if 'video' in jingle_types:
            if self.jingle['video'].state != JingleState.NULL:
                self.close_jingle_content('video')
            else:
                sid = con.get_module('Jingle').start_video(
                    self.contact.get_full_jid())
                self.set_jingle_state('video', JingleState.CONNECTING, sid)

    def _get_audio_content(self):
        con = app.connections[self.account]
        session = con.get_module('Jingle').get_jingle_session(
            self.contact.get_full_jid(), self.jingle['audio'].sid)
        return session.get_content('audio')

    def on_num_button_pressed(self, _widget, num):
        self._get_audio_content().start_dtmf(num)

    def on_num_button_released(self, _released):
        self._get_audio_content().stop_dtmf()

    def on_mic_hscale_value_changed(self, _widget, value):
        self._get_audio_content().set_mic_volume(value / 100)
        app.settings.set('audio_input_volume', int(value))

    def on_sound_hscale_value_changed(self, _widget, value):
        self._get_audio_content().set_out_volume(value / 100)
        app.settings.set('audio_output_volume', int(value))

    def on_location_eventbox_button_release_event(self, _widget, _event):
        if 'geoloc' in self.contact.pep:
            location = self.contact.pep['geoloc'].data
            if 'lat' in location and 'lon' in location:
                uri = geo_provider_from_location(location['lat'],
                                                 location['lon'])
                open_uri(uri)

    def on_location_eventbox_leave_notify_event(self, _widget, _event):
        """
        Just moved the mouse so show the cursor
        """
        cursor = get_cursor('default')
        self.parent_win.window.get_window().set_cursor(cursor)

    def on_location_eventbox_enter_notify_event(self, _widget, _event):
        cursor = get_cursor('pointer')
        self.parent_win.window.get_window().set_cursor(cursor)

    def update_ui(self):
        # The name banner is drawn here
        ChatControlBase.update_ui(self)
        self.update_toolbar()
        self._update_avatar()
        self.update_actions()

    def draw_banner_text(self):
        """
        Draw the text in the fat line at the top of the window that houses the
        name, jid
        """
        contact = self.contact
        name = contact.get_shown_name()
        if self.resource:
            name += '/' + self.resource
        if self._type.is_privatechat:
            name = i18n.direction_mark + _(
                '%(nickname)s from group chat %(room_name)s') % \
                {'nickname': name, 'room_name': self.room_name}
        name = i18n.direction_mark + GLib.markup_escape_text(name)

        status = contact.status
        if status is not None:
            status_reduced = helpers.reduce_chars_newlines(status, max_lines=1)
        else:
            status_reduced = ''
        status_escaped = GLib.markup_escape_text(status_reduced)

        if self._type.is_privatechat:
            cs = self.gc_contact.chatstate
        else:
            cs = app.contacts.get_combined_chatstate(
                self.account, self.contact.jid)

        if app.settings.get('show_chatstate_in_banner'):
            chatstate = helpers.get_uf_chatstate(cs)

            label_text = '<span>%s</span><span size="x-small" weight="light"> %s</span>' % \
                (name, chatstate)
            label_tooltip = '%s %s' % (name, chatstate)
        else:
            label_text = '<span>%s</span>' % name
            label_tooltip = name

        if status_escaped:
            status_text = make_href_markup(status_escaped)
            status_text = '<span size="x-small" weight="light">%s</span>' % status_text
            self.xml.banner_label.set_tooltip_text(status)
            self.xml.banner_label.set_no_show_all(False)
            self.xml.banner_label.show()
        else:
            status_text = ''
            self.xml.banner_label.hide()
            self.xml.banner_label.set_no_show_all(True)

        self.xml.banner_label.set_markup(status_text)
        # setup the label that holds name and jid
        self.xml.banner_name_label.set_markup(label_text)
        self.xml.banner_name_label.set_tooltip_text(label_tooltip)

    def send_message(self,
                     message,
                     xhtml=None,
                     process_commands=True,
                     attention=False):
        """
        Send a message to contact
        """

        if self.encryption:
            self.sendmessage = True
            app.plugin_manager.extension_point('send_message' + self.encryption,
                                               self)
            if not self.sendmessage:
                return

        message = helpers.remove_invalid_xml_chars(message)
        if message in ('', None, '\n'):
            return

        ChatControlBase.send_message(self,
                                     message,
                                     type_='chat',
                                     xhtml=xhtml,
                                     process_commands=process_commands,
                                     attention=attention)

    def get_our_nick(self):
        return app.nicks[self.account]

    def add_message(self,
                    text,
                    frm='',
                    tim=None,
                    subject=None,
                    displaymarking=None,
                    msg_log_id=None,
                    correct_id=None,
                    message_id=None,
                    additional_data=None,
                    error=None):
        """
        Print a line in the conversation

        If frm is set to status: it's a status message.
        if frm is set to error: it's an error message. The difference between
            status and error is mainly that with error, msg count as a new
            message (in systray and in control).
        If frm is set to info: it's a information message.
        If frm is set to print_queue: it is incoming from queue.
        If frm is set to another value: it's an outgoing message.
        If frm is not set: it's an incoming message.
        """
        contact = self.contact

        if additional_data is None:
            additional_data = AdditionalDataDict()

        if frm == 'error':
            kind = 'error'
            name = ''
        else:
            if not frm:
                kind = 'incoming'
                name = contact.get_shown_name()
            elif frm == 'print_queue':
                kind = 'incoming_queue'
                name = contact.get_shown_name()
            else:
                kind = 'outgoing'
                name = self.get_our_nick()

        ChatControlBase.add_message(self,
                                    text,
                                    kind,
                                    name,
                                    tim,
                                    subject=subject,
                                    old_kind=self.old_msg_kind,
                                    displaymarking=displaymarking,
                                    msg_log_id=msg_log_id,
                                    message_id=message_id,
                                    correct_id=correct_id,
                                    additional_data=additional_data,
                                    error=error)
        if text.startswith('/me ') or text.startswith('/me\n'):
            self.old_msg_kind = None
        else:
            self.old_msg_kind = kind

    def get_tab_label(self):
        unread = ''
        if self.resource:
            jid = self.contact.get_full_jid()
        else:
            jid = self.contact.jid
        num_unread = len(app.events.get_events(
            self.account, jid, ['printed_%s' % self._type, str(self._type)]))
        if num_unread == 1:
            unread = '*'
        elif num_unread > 1:
            unread = '[' + str(num_unread) + ']'

        name = self.contact.get_shown_name()
        if self.resource:
            name += '/' + self.resource
        label_str = GLib.markup_escape_text(name)
        if num_unread: # if unread, text in the label becomes bold
            label_str = '<b>' + unread + label_str + '</b>'
        return label_str

    def get_tab_image(self):
        scale = self.parent_win.window.get_scale_factor()
        return app.contacts.get_avatar(self.account,
                                       self.contact.jid,
                                       AvatarSize.ROSTER,
                                       scale,
                                       self.contact.show)

    def prepare_context_menu(self, hide_buttonbar_items=False):
        """
        Set compact view menuitem active state sets active and sensitivity state
        for history_menuitem (False for tranasports) and file_transfer_menuitem
        and hide()/show() for add_to_roster_menuitem
        """
        if app.jid_is_transport(self.contact.jid):
            menu = gui_menu_builder.get_transport_menu(self.contact,
                                                       self.account)
        else:
            menu = gui_menu_builder.get_contact_menu(
                self.contact,
                self.account,
                use_multiple_contacts=False,
                show_start_chat=False,
                show_encryption=True,
                control=self,
                show_buttonbar_items=not hide_buttonbar_items)
        return menu

    def shutdown(self):
        # PluginSystem: removing GUI extension points connected with ChatControl
        # instance object
        app.plugin_manager.remove_gui_extension_point('chat_control', self)

        self.unsubscribe_events()

        self.remove_actions()

        # Send 'gone' chatstate
        con = app.connections[self.account]
        con.get_module('Chatstate').set_chatstate(self.contact, Chatstate.GONE)

        for jingle_type in ('audio', 'video'):
            self.close_jingle_content(jingle_type, shutdown=True)
        self.jingle.clear()

        # disconnect self from session
        if self.session:
            self.session.control = None

        # Clean events
        app.events.remove_events(
            self.account,
            self.get_full_jid(),
            types=['printed_%s' % self._type, str(self._type)])
        # Remove contact instance if contact has been removed
        key = (self.contact.jid, self.account)
        roster = app.interface.roster
        has_pending = roster.contact_has_pending_roster_events(self.contact,
                                                               self.account)
        if key in roster.contacts_to_be_removed.keys() and not has_pending:
            backend = roster.contacts_to_be_removed[key]['backend']
            del roster.contacts_to_be_removed[key]
            roster.remove_contact(self.contact.jid,
                                  self.account,
                                  force=True,
                                  backend=backend)

        super(ChatControl, self).shutdown()

    def minimizable(self):
        return False

    def safe_shutdown(self):
        return False

    def allow_shutdown(self, method, on_yes, on_no, _on_minimize):
        time_ = app.last_message_time[self.account][self.get_full_jid()]
        # 2 seconds
        if time.time() - time_ < 2:
            no_log_for = app.settings.get_account_setting(
                self.account, 'no_log_for').split()
            more = ''
            if self.contact.jid in no_log_for:
                more = _('Note: Chat history is disabled for this contact.')
            if self.account in no_log_for:
                more = _('Note: Chat history is disabled for this account.')
            text = _('You just received a new message from %s.\n'
                     'Do you want to close this tab?') % self.contact.get_shown_name()
            if more:
                text += '\n' + more

            ConfirmationDialog(
                _('Close'),
                _('New Message'),
                text,
                [DialogButton.make('Cancel',
                                   callback=lambda: on_no(self)),
                 DialogButton.make('Remove',
                                   text=_('_Close'),
                                   callback=lambda: on_yes(self))],
                transient_for=self.parent_win.window).show()
            return
        on_yes(self)

    def _update_avatar(self):
        scale = self.parent_win.window.get_scale_factor()
        surface = app.contacts.get_avatar(self.account,
                                          self.contact.jid,
                                          AvatarSize.CHAT,
                                          scale,
                                          self.contact.show)

        self.xml.avatar_image.set_from_surface(surface)

    def _on_drag_data_received(self, widget, context, x, y, selection,
                               target_type, timestamp):
        if not selection.get_data():
            return

        if target_type == self.TARGET_TYPE_URI_LIST:
            # File drag and drop (handled in chat_control_base)
            self.drag_data_file_transfer(selection)
        else:
            # Convert single chat to MUC
            treeview = app.interface.roster.tree
            model = treeview.get_model()
            data = selection.get_data().decode()
            tree_selection = treeview.get_selection()
            if tree_selection.count_selected_rows() == 0:
                return
            path = tree_selection.get_selected_rows()[1][0]
            iter_ = model.get_iter(path)
            type_ = model[iter_][2]
            if type_ != 'contact':  # Source is not a contact
                return
            dropped_jid = data

            dropped_transport = app.get_transport_name_from_jid(dropped_jid)
            c_transport = app.get_transport_name_from_jid(self.contact.jid)
            if dropped_transport or c_transport:
                return # transport contacts cannot be invited

            dialogs.TransformChatToMUC(self.account,
                                       [self.contact.jid],
                                       [dropped_jid])

    def restore_conversation(self):
        jid = self.contact.jid
        # don't restore lines if it's a transport
        if app.jid_is_transport(jid):
            return

        # number of messages that are in queue and are already logged, we want
        # to avoid duplication
        pending = len(app.events.get_events(self.account, jid, ['chat', 'pm']))
        if self.resource:
            pending += len(app.events.get_events(self.account,
                                                 self.contact.get_full_jid(),
                                                 ['chat', 'pm']))

        rows = app.storage.archive.get_last_conversation_lines(
            self.account, jid, pending)

        local_old_kind = None
        self.conv_textview.just_cleared = True
        for row in rows: # time, kind, message, subject, additional_data
            msg = row.message
            additional_data = row.additional_data
            if not msg: # message is empty, we don't print it
                continue
            if row.kind in (KindConstant.CHAT_MSG_SENT,
                            KindConstant.SINGLE_MSG_SENT):
                kind = 'outgoing'
                name = self.get_our_nick()
            elif row.kind in (KindConstant.SINGLE_MSG_RECV,
                              KindConstant.CHAT_MSG_RECV):
                kind = 'incoming'
                name = self.contact.get_shown_name()
            elif row.kind == KindConstant.ERROR:
                kind = 'status'
                name = self.contact.get_shown_name()

            tim = float(row.time)

            if row.subject:
                msg = _('Subject: %(subject)s\n%(message)s') % \
                    {'subject': row.subject, 'message': msg}
            ChatControlBase.add_message(self,
                                        msg,
                                        kind,
                                        name,
                                        tim,
                                        restored=True,
                                        old_kind=local_old_kind,
                                        additional_data=additional_data,
                                        message_id=row.message_id,
                                        marker=row.marker,
                                        error=row.error)
            if (row.message.startswith('/me ') or
                    row.message.startswith('/me\n')):
                local_old_kind = None
            else:
                local_old_kind = kind
        if rows:
            self.conv_textview.print_empty_line()

    def read_queue(self):
        """
        Read queue and print messages contained in it
        """
        jid = self.contact.jid
        jid_with_resource = jid
        if self.resource:
            jid_with_resource += '/' + self.resource
        events = app.events.get_events(self.account, jid_with_resource)

        # list of message ids which should be marked as read
        message_ids = []
        for event in events:
            if event.type_ != str(self._type):
                continue
            kind = 'print_queue'
            if event.sent_forwarded:
                kind = 'out'
            self.add_message(event.message,
                             kind,
                             tim=event.time,
                             subject=event.subject,
                             displaymarking=event.displaymarking,
                             correct_id=event.correct_id,
                             message_id=event.message_id,
                             additional_data=event.additional_data)
            if isinstance(event.msg_log_id, int):
                message_ids.append(event.msg_log_id)

            if event.session and not self.session:
                self.set_session(event.session)
        if message_ids:
            app.storage.archive.set_read_messages(message_ids)

        # XEP-0333 Send <displayed> marker
        con = app.connections[self.account]
        con.get_module('ChatMarkers').send_displayed_marker(
            self.contact,
            self.last_msg_id,
            self._type)
        self.last_msg_id = None
        app.events.remove_events(self.account,
                                 jid_with_resource,
                                 types=[str(self._type)])

        typ = 'chat' # Is it a normal chat or a pm ?

        # reset to status image in gc if it is a pm
        # Is it a pm ?
        room_jid, nick = app.get_room_and_nick_from_fjid(jid)
        control = app.interface.msg_win_mgr.get_gc_control(room_jid,
                                                           self.account)
        if control and control.is_groupchat:
            control.update_ui()
            control.parent_win.show_title()
            typ = 'pm'

        self.redraw_after_event_removed(jid)
        if self.contact.show in ('offline', 'error'):
            show_offline = app.settings.get('showoffline')
            show_transports = app.settings.get('show_transports_group')
            if (not show_transports and app.jid_is_transport(jid)) or \
            (not show_offline and typ == 'chat' and \
            len(app.contacts.get_contacts(self.account, jid)) < 2):
                app.interface.roster.remove_to_be_removed(self.contact.jid,
                                                          self.account)
            elif typ == 'pm':
                control.remove_contact(nick)

    def _on_convert_to_gc_menuitem_activate(self, _widget):
        """
        User wants to invite some friends to chat
        """
        dialogs.TransformChatToMUC(self.account, [self.contact.jid])

    def got_connected(self):
        ChatControlBase.got_connected(self)
        # Refreshing contact
        contact = app.contacts.get_contact_with_highest_priority(
            self.account, self.contact.jid)
        if isinstance(contact, GC_Contact):
            contact = contact.as_contact()
        if contact:
            self.contact = contact
        self.draw_banner()
        self.update_actions()

    def got_disconnected(self):
        ChatControlBase.got_disconnected(self)
        self.update_actions()

    def update_status_display(self, name, uf_show, status):
        self.update_ui()
        self.parent_win.redraw_tab(self)

        if not app.settings.get('print_status_in_chats'):
            return

        if status:
            status = '- %s' % status
        status_line = _('%(name)s is now %(show)s %(status)s') % {
            'name': name, 'show': uf_show, 'status': status or ''}
        self.add_status_message(status_line)

    def _info_bar_show_message(self):
        if self.info_bar.get_visible():
            # A message is already shown
            return
        if not self.info_bar_queue:
            return
        markup, buttons, _args, type_ = self.info_bar_queue[0]
        self.info_bar_label.set_markup(markup)

        # Remove old buttons
        area = self.info_bar.get_action_area()
        for button in area.get_children():
            area.remove(button)

        # Add new buttons
        for button in buttons:
            self.info_bar.add_action_widget(button, 0)

        self.info_bar.set_message_type(type_)
        self.info_bar.set_no_show_all(False)
        self.info_bar.show_all()

    def _add_info_bar_message(self, markup, buttons, args,
                              type_=Gtk.MessageType.INFO):
        self.info_bar_queue.append((markup, buttons, args, type_))
        self._info_bar_show_message()

    def _get_file_props_event(self, file_props, type_):
        evs = app.events.get_events(self.account, self.contact.jid, [type_])
        for ev in evs:
            if ev.file_props == file_props:
                return ev
        return None

    def _on_accept_file_request(self, _widget, file_props):
        app.interface.instances['file_transfers'].on_file_request_accepted(
            self.account, self.contact, file_props)
        ev = self._get_file_props_event(file_props, 'file-request')
        if ev:
            app.events.remove_events(self.account, self.contact.jid, event=ev)

    def _on_cancel_file_request(self, _widget, file_props):
        con = app.connections[self.account]
        con.get_module('Bytestream').send_file_rejection(file_props)
        ev = self._get_file_props_event(file_props, 'file-request')
        if ev:
            app.events.remove_events(self.account, self.contact.jid, event=ev)

    def _got_file_request(self, file_props):
        """
        Show an InfoBar on top of control
        """
        if app.settings.get('use_kib_mib'):
            units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            units = GLib.FormatSizeFlags.DEFAULT

        markup = '<b>%s</b>\n%s' % (_('File Transfer'), file_props.name)
        if file_props.desc:
            markup += '\n(%s)' % file_props.desc
        markup += '\n%s: %s' % (
            _('Size'),
            GLib.format_size_full(file_props.size, units))
        button_decline = Gtk.Button.new_with_mnemonic(_('_Decline'))
        button_decline.connect(
            'clicked', self._on_cancel_file_request, file_props)
        button_accept = Gtk.Button.new_with_mnemonic(_('_Accept'))
        button_accept.connect(
            'clicked', self._on_accept_file_request, file_props)
        self._add_info_bar_message(
            markup,
            [button_decline, button_accept],
            file_props,
            Gtk.MessageType.QUESTION)

    def _on_open_ft_folder(self, _widget, file_props):
        path = os.path.split(file_props.file_name)[0]
        if os.path.exists(path) and os.path.isdir(path):
            open_file(path)
        ev = self._get_file_props_event(file_props, 'file-completed')
        if ev:
            app.events.remove_events(self.account, self.contact.jid, event=ev)

    def _on_ok(self, _widget, file_props, type_):
        ev = self._get_file_props_event(file_props, type_)
        if ev:
            app.events.remove_events(self.account, self.contact.jid, event=ev)

    def _got_file_completed(self, file_props):
        markup = '<b>%s</b>\n%s' % (_('File Transfer Completed'),
                                    file_props.name)
        if file_props.desc:
            markup += '\n(%s)' % file_props.desc
        b1 = Gtk.Button.new_with_mnemonic(_('Open _Folder'))
        b1.connect('clicked', self._on_open_ft_folder, file_props)
        b2 = Gtk.Button.new_with_mnemonic(_('_Close'))
        b2.connect('clicked', self._on_ok, file_props, 'file-completed')
        self._add_info_bar_message(
            markup,
            [b1, b2],
            file_props)

    def _got_file_error(self, file_props, type_, pri_txt, sec_txt):
        markup = '<b>%s</b>\n%s' % (pri_txt, sec_txt)
        button = Gtk.Button.new_with_mnemonic(_('_Close'))
        button.connect('clicked', self._on_ok, file_props, type_)
        self._add_info_bar_message(
            markup,
            [button],
            file_props,
            Gtk.MessageType.ERROR)

    def _on_accept_gc_invitation(self, _widget, event):
        app.interface.show_or_join_groupchat(self.account,
                                             str(event.muc),
                                             password=event.password)
        app.events.remove_events(self.account, self.contact.jid, event=event)

    def _on_cancel_gc_invitation(self, _widget, event):
        app.events.remove_events(self.account, self.contact.jid, event=event)

    def _get_gc_invitation(self, event):
        markup = '<b>%s</b>\n%s' % (_('Group Chat Invitation'), event.muc)
        if event.reason:
            markup += '\n(%s)' % event.reason
        button_decline = Gtk.Button.new_with_mnemonic(_('_Decline'))
        button_decline.connect('clicked', self._on_cancel_gc_invitation, event)
        button_accept = Gtk.Button.new_with_mnemonic(_('_Accept'))
        button_accept.connect('clicked', self._on_accept_gc_invitation, event)
        self._add_info_bar_message(
            markup,
            [button_decline, button_accept],
            (event.muc, event.reason),
            Gtk.MessageType.QUESTION)

    def _on_reject_call(self, _button, event):
        app.events.remove_events(
            self.account, self.contact.jid, types='jingle-incoming')

        con = app.connections[self.account]
        session = con.get_module('Jingle').get_jingle_session(
            event.peerjid, event.sid)
        if not session:
            return

        if not session.accepted:
            session.decline_session()
        else:
            for content in event.content_types:
                session.reject_content(content)

    def _on_accept_call(self, _button, event):
        app.events.remove_events(
            self.account, self.contact.jid, types='jingle-incoming')

        con = app.connections[self.account]
        session = con.get_module('Jingle').get_jingle_session(
            event.peerjid, event.sid)
        if not session:
            return

        audio = session.get_content('audio')
        video = session.get_content('video')

        if audio and not audio.negotiated:
            self.set_jingle_state('audio', JingleState.CONNECTING, event.sid)
        if video and not video.negotiated:
            self.set_jingle_state('video', JingleState.CONNECTING, event.sid)

        if not session.accepted:
            session.approve_session()

        for content in event.content_types:
            session.approve_content(content)

    def add_call_received_message(self, event):
        markup = '<b>%s</b>' % (_('Incoming Call'))
        if 'video' in event.content_types:
            markup += _('\nVideo Call')
        else:
            markup += _('\nVoice Call')

        button_reject = Gtk.Button.new_with_mnemonic(_('_Reject'))
        button_reject.connect('clicked', self._on_reject_call, event)
        button_accept = Gtk.Button.new_with_mnemonic(_('_Accept'))
        button_accept.connect('clicked', self._on_accept_call, event)
        self._add_info_bar_message(
            markup,
            [button_reject, button_accept],
            event,
            Gtk.MessageType.QUESTION)

    def on_event_added(self, event):
        if event.account != self.account:
            return
        if event.jid != self.contact.jid:
            return
        if event.type_ == 'file-request':
            self._got_file_request(event.file_props)
        elif event.type_ == 'file-completed':
            self._got_file_completed(event.file_props)
        elif event.type_ in ('file-error', 'file-stopped'):
            msg_err = ''
            if event.file_props.error == -1:
                msg_err = _('Remote contact stopped transfer')
            elif event.file_props.error == -6:
                msg_err = _('Error opening file')
            self._got_file_error(event.file_props, event.type_,
                                 _('File transfer stopped'), msg_err)
        elif event.type_ in ('file-request-error', 'file-send-error'):
            self._got_file_error(
                event.file_props,
                event.type_,
                _('File transfer cancelled'),
                _('Connection with peer cannot be established.'))
        elif event.type_ == 'gc-invitation':
            self._get_gc_invitation(event)

    def on_event_removed(self, event_list):
        """
        Called when one or more events are removed from the event list
        """
        for ev in event_list:
            if ev.account != self.account:
                continue
            if ev.jid != self.contact.jid:
                continue
            if ev.type_ not in ('file-request',
                                'file-completed',
                                'file-error',
                                'file-stopped',
                                'file-request-error',
                                'file-send-error',
                                'gc-invitation',
                                'jingle-incoming'):
                continue
            i = 0
            removed = False
            for ib_msg in self.info_bar_queue:
                if ev.type_ == 'gc-invitation':
                    if ev.muc == ib_msg[2][0]:
                        self.info_bar_queue.remove(ib_msg)
                        removed = True
                elif ev.type_ == 'jingle-incoming':
                    # TODO: Need to be more specific here?
                    self.info_bar_queue.remove(ib_msg)
                    removed = True
                else: # file-*
                    if ib_msg[2] == ev.file_props:
                        self.info_bar_queue.remove(ib_msg)
                        removed = True
                if removed:
                    if i == 0:
                        # We are removing the one currently displayed
                        self.info_bar.set_no_show_all(True)
                        self.info_bar.hide()
                        # show next one?
                        GLib.idle_add(self._info_bar_show_message)
                    break
                i += 1



# Copyright (C) 2006 Jean-Marie Traissard <jim AT lapin.org>
#                    Junglecow J <junglecow AT gmail.com>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Brendan Taylor <whateley AT gmail.com>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim. If not, see <http://www.gnu.org/licenses/>.

from typing import Dict  # pylint: disable=unused-import
from typing import List
from typing import Generator
from typing import Optional  # pylint: disable=unused-import
from typing import Tuple

import os
import sys
import tempfile
from pathlib import Path

from gi.repository import GLib

import gajim
from gajim.common.i18n import _
from gajim.common.const import PathType, PathLocation
from gajim.common.types import PathTuple


def get(key: str) -> Path:
    return _paths[key]


def get_plugin_dirs() -> List[Path]:
    if gajim.IS_FLATPAK:
        return [Path(_paths['PLUGINS_BASE']),
                Path('/app/plugins')]
    return [Path(_paths['PLUGINS_BASE']),
            Path(_paths['PLUGINS_USER'])]


def get_paths(type_: PathType) -> Generator[Path, None, None]:
    for key, value in _paths.items():
        path_type = value[2]
        if type_ != path_type:
            continue
        yield _paths[key]


def override_path(*args, **kwargs):
    _paths.add(*args, **kwargs)


def set_separation(active: bool) -> None:
    _paths.profile_separation = active


def set_profile(profile: str) -> None:
    _paths.profile = profile


def set_config_root(config_root: str) -> None:
    _paths.custom_config_root = Path(config_root).resolve()


def init() -> None:
    _paths.init()


def create_paths() -> None:
    for path in get_paths(PathType.FOLDER):
        if path.is_file():
            print(_('%s is a file but it should be a directory') % path)
            print(_('Gajim will now exit'))
            sys.exit()

        if not path.exists():
            for parent_path in reversed(path.parents):
                # Create all parent folders
                # don't use mkdir(parent=True), as it ignores `mode`
                # when creating the parents
                if not parent_path.exists():
                    print(('creating %s directory') % parent_path)
                    parent_path.mkdir(mode=0o700)
            print(('creating %s directory') % path)
            path.mkdir(mode=0o700)


class ConfigPaths:
    def __init__(self) -> None:
        self._paths = {}  # type: Dict[str, PathTuple]
        self.profile = ''
        self.profile_separation = False
        self.custom_config_root = None  # type: Optional[Path]

        if os.name == 'nt':
            if gajim.IS_PORTABLE:
                application_path = Path(sys.executable).parent
                self.config_root = self.cache_root = self.data_root = \
                        application_path.parent / 'UserData'
            else:
                # Documents and Settings\[User Name]\Application Data\Gajim
                self.config_root = self.cache_root = self.data_root = \
                        Path(os.environ['appdata']) / 'Gajim'
        else:
            self.config_root = Path(GLib.get_user_config_dir()) / 'gajim'
            self.cache_root = Path(GLib.get_user_cache_dir()) / 'gajim'
            self.data_root = Path(GLib.get_user_data_dir()) / 'gajim'

        import pkg_resources
        basedir = Path(pkg_resources.resource_filename("gajim", "."))

        source_paths = [
            ('DATA', basedir / 'data'),
            ('STYLE', basedir / 'data' / 'style'),
            ('EMOTICONS', basedir / 'data' / 'emoticons'),
            ('GUI', basedir / 'data' / 'gui'),
            ('ICONS', basedir / 'data' / 'icons'),
            ('HOME', Path.home()),
            ('PLUGINS_BASE', basedir / 'data' / 'plugins'),
        ]

        for path in source_paths:
            self.add(*path)

    def __getitem__(self, key: str) -> Path:
        location, path, _ = self._paths[key]
        if location == PathLocation.CONFIG:
            return self.config_root / path
        if location == PathLocation.CACHE:
            return self.cache_root / path
        if location == PathLocation.DATA:
            return self.data_root / path
        return path

    def items(self) -> Generator[Tuple[str, PathTuple], None, None]:
        for key, value in self._paths.items():
            yield (key, value)

    def _prepare(self, path: Path, unique: bool) -> Path:
        if os.name == 'nt':
            path = Path(str(path).capitalize())
        if self.profile:
            if unique or self.profile_separation:
                return Path(f'{path}.{self.profile}')
        return path

    def add(self,
            name: str,
            path: Path,
            location: PathLocation = None,
            path_type: PathType = None,
            unique: bool = False) -> None:
        if location is not None:
            path = self._prepare(path, unique)
        self._paths[name] = (location, path, path_type)

    def init(self):
        if self.custom_config_root:
            self.config_root = self.custom_config_root
            self.cache_root = self.data_root = self.custom_config_root

        user_dir_paths = [
            ('TMP', Path(tempfile.gettempdir())),
            ('MY_CONFIG', Path(), PathLocation.CONFIG, PathType.FOLDER),
            ('MY_CACHE', Path(), PathLocation.CACHE, PathType.FOLDER),
            ('MY_DATA', Path(), PathLocation.DATA, PathType.FOLDER),
        ]

        for path in user_dir_paths:
            self.add(*path)

        # These paths are unique per profile
        unique_profile_paths = [
            # Data paths
            ('SECRETS_FILE', 'secrets', PathLocation.DATA, PathType.FILE),
            ('MY_PEER_CERTS', 'certs', PathLocation.DATA, PathType.FOLDER),
            ('CERT_STORE', 'cert_store', PathLocation.DATA, PathType.FOLDER),
            ('DEBUG', 'debug', PathLocation.DATA, PathType.FOLDER),
            ('PLUGINS_DATA', 'plugins_data', PathLocation.DATA, PathType.FOLDER),

            # Config paths
            ('SETTINGS', 'settings.sqlite', PathLocation.CONFIG, PathType.FILE),
            ('CONFIG_FILE', 'config', PathLocation.CONFIG, PathType.FILE),
            ('PLUGINS_CONFIG_DIR',
             'pluginsconfig', PathLocation.CONFIG, PathType.FOLDER),
            ('MY_CERT', 'localcerts', PathLocation.CONFIG, PathType.FOLDER),
        ]

        for path in unique_profile_paths:
            self.add(*path, unique=True)

        # These paths are only unique per profile if the commandline arg
        # `separate` is passed
        paths = [
            # Data paths
            ('LOG_DB', 'logs.db', PathLocation.DATA, PathType.FILE),
            ('MY_CACERTS', 'cacerts.pem', PathLocation.DATA, PathType.FILE),
            ('PLUGINS_DOWNLOAD', 'plugins_download', PathLocation.CACHE, PathType.FOLDER),
            ('PLUGINS_USER', 'plugins', PathLocation.DATA, PathType.FOLDER),
            ('MY_EMOTS',
             'emoticons', PathLocation.DATA, PathType.FOLDER_OPTIONAL),
            ('MY_ICONSETS',
             'iconsets', PathLocation.DATA, PathType.FOLDER_OPTIONAL),

            # Cache paths
            ('CACHE_DB', 'cache.db', PathLocation.CACHE, PathType.FILE),
            ('AVATAR', 'avatars', PathLocation.CACHE, PathType.FOLDER),
            ('BOB', 'bob', PathLocation.CACHE, PathType.FOLDER),

            # Config paths
            ('MY_THEME', 'theme', PathLocation.CONFIG, PathType.FOLDER),

        ]

        for path in paths:
            self.add(*path)


_paths = ConfigPaths()



# Copyright (C) 2009 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim. If not, see <http://www.gnu.org/licenses/>.

# XEP-0115: Entity Capabilities

import weakref
from collections import defaultdict

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import DiscoIdentity
from nbxmpp.util import compute_caps_hash
from nbxmpp.errors import StanzaError

from gajim.common import app
from gajim.common.const import COMMON_FEATURES
from gajim.common.const import Entity
from gajim.common.helpers import get_optional_features
from gajim.common.nec import NetworkEvent
from gajim.common.task_manager import Task
from gajim.common.modules.base import BaseModule


class Caps(BaseModule):

    _nbxmpp_extends = 'EntityCaps'
    _nbxmpp_methods = [
        'caps',
        'set_caps'
    ]

    def __init__(self, con):
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='presence',
                          callback=self._entity_caps,
                          ns=Namespace.CAPS,
                          priority=51),
        ]

        self._identities = [
            DiscoIdentity(category='client', type='pc', name='Gajim')
        ]

        self._queued_tasks_by_hash = defaultdict(set)
        self._queued_tasks_by_jid = {}

    def _queue_task(self, task):
        old_task = self._get_task(task.entity.jid)
        if old_task is not None:
            self._remove_task(old_task)

        self._log.info('Queue query for hash %s', task.entity.hash)
        self._queued_tasks_by_hash[task.entity.hash].add(task)
        self._queued_tasks_by_jid[task.entity.jid] = task
        app.task_manager.add_task(task)

    def _get_task(self, jid):
        return self._queued_tasks_by_jid.get(jid)

    def _get_similar_tasks(self, task):
        return self._queued_tasks_by_hash.pop(task.entity.hash)

    def _remove_task(self, task):
        task.set_obsolete()
        del self._queued_tasks_by_jid[task.entity.jid]
        self._queued_tasks_by_hash[task.entity.hash].discard(task)

    def _remove_all_tasks(self):
        for task in self._queued_tasks_by_jid.values():
            task.set_obsolete()
        self._queued_tasks_by_jid.clear()
        self._queued_tasks_by_hash.clear()

    def _entity_caps(self, _con, _stanza, properties):
        if properties.type.is_error or properties.type.is_unavailable:
            return

        if properties.is_self_presence:
            return

        if properties.entity_caps is None:
            return

        task = EntityCapsTask(self._account, properties, self._execute_task)

        self._log.info('Received %s', task.entity)

        disco_info = app.storage.cache.get_caps_entry(task.entity.method,
                                                      task.entity.hash)
        if disco_info is None:
            self._queue_task(task)
            return

        jid = str(properties.jid)
        app.storage.cache.set_last_disco_info(jid, disco_info, cache_only=True)
        app.nec.push_incoming_event(
            NetworkEvent('caps-update',
                         account=self._account,
                         fjid=jid,
                         jid=properties.jid.bare))

    def _execute_task(self, task):
        self._log.info('Request %s from %s', task.entity.hash, task.entity.jid)
        self._con.get_module('Discovery').disco_info(
            task.entity.jid,
            node=f'{task.entity.node}#{task.entity.hash}',
            callback=self._on_disco_info,
            user_data=task.entity.jid)

    def _on_disco_info(self, nbxmpp_task):
        jid = nbxmpp_task.get_user_data()
        task = self._get_task(jid)
        if task is None:
            self._log.info('Task not found for %s', jid)
            return

        self._remove_task(task)

        try:
            disco_info = nbxmpp_task.finish()
        except StanzaError as error:
            self._log.warning(error)
            return

        self._log.info('Disco Info received: %s', disco_info.jid)

        try:
            compute_caps_hash(disco_info)
        except Exception as error:
            self._log.warning('Disco info malformed: %s %s',
                              disco_info.jid, error)
            return

        app.storage.cache.add_caps_entry(
            str(disco_info.jid),
            task.entity.method,
            disco_info.get_caps_hash(),
            disco_info)

        self._log.info('Finished query for %s', task.entity.hash)

        tasks = self._get_similar_tasks(task)

        for task in tasks:
            self._remove_task(task)
            self._log.info('Update %s', task.entity.jid)
            app.nec.push_incoming_event(
                NetworkEvent('caps-update',
                             account=self._account,
                             fjid=str(task.entity.jid),
                             jid=task.entity.jid.bare))

    def update_caps(self):
        if not app.account_is_connected(self._account):
            return

        optional_features = get_optional_features(self._account)
        self.set_caps(self._identities,
                      COMMON_FEATURES + optional_features,
                      'https://gajim.org')

        if not app.account_is_available(self._account):
            return

        app.connections[self._account].change_status(
            app.connections[self._account].status,
            app.connections[self._account].status_message)

    def cleanup(self):
        self._remove_all_tasks()
        BaseModule.cleanup(self)


class EntityCapsTask(Task):
    def __init__(self, account, properties, callback):
        Task.__init__(self)
        self._account = account
        self._callback = weakref.WeakMethod(callback)

        self.entity = Entity(jid=properties.jid,
                             node=properties.entity_caps.node,
                             hash=properties.entity_caps.ver,
                             method=properties.entity_caps.hash)

        self._from_muc = properties.from_muc

    def execute(self):
        callback = self._callback()
        if callback is not None:
            callback(self)

    def preconditions_met(self):
        try:
            client = app.get_client(self._account)
        except Exception:
            return False

        if self._from_muc:
            muc = client.get_module('MUC').get_manager().get(
                self.entity.jid.bare)

            if muc is None or not muc.state.is_joined:
                self.set_obsolete()
                return False

        return client.state.is_available

    def __repr__(self):
        return f'Entity Caps ({self.entity.jid} {self.entity.hash})'

    def __hash__(self):
        return hash(self.entity)


def get_instance(*args, **kwargs):
    return Caps(*args, **kwargs), 'Caps'



import time

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from nbxmpp.protocol import Iq
from nbxmpp.modules.discovery import parse_disco_info

from gajim.common.const import CSSPriority

from gajim import gui
gui.init('gtk')

from test.gtk import util
from gajim.gui.groupchat_info import GroupChatInfoScrolled

util.load_style('gajim.css', CSSPriority.APPLICATION)

stanza = Iq(node='''
<iq xmlns="jabber:client" xml:lang="de-DE" to="user@user.us" from="asd@conference.temptatio.dev" type="result" id="67284933-e526-41f3-8309-9d9475cf9c74">
<query xmlns="http://jabber.org/protocol/disco#info">
<identity name="ipsum dolor sit amet, consetetur sadipscing elitr sed diam nonumy eirmod tempor invidunt" type="text" category="conference" />
<feature var="vcard-temp" />
<feature var="http://jabber.org/protocol/muc" />
<feature var="http://jabber.org/protocol/disco#info" />
<feature var="http://jabber.org/protocol/disco#items" />
<feature var="muc_temporary" />
<feature var="muc_moderated" />
<feature var="muc_open" />
<feature var="muc_hidden" />
<feature var="muc_nonanonymous" />
<feature var="muc_passwordprotected" />
<feature var="urn:xmpp:mam:2" />
<feature var="muc_public" />
<feature var="muc_persistent" />
<feature var="muc_membersonly" />
<feature var="muc_semianonymous" />
<feature var="muc_unmoderated" />
<feature var="muc_unsecured" />
<x type="result" xmlns="jabber:x:data">
<field var="FORM_TYPE" type="hidden">
<value>http://jabber.org/protocol/muc#roominfo</value>
</field>
<field var="muc#roominfo_occupants" type="text-single" label="Number of occupants">
<value>1</value>
</field>
<field var="muc#roomconfig_roomname" type="text-single" label="Natural-Language Room Name">
<value>ipsum dolor sit amet, consetetur sadipscing elitr sed diam nonumy eirmod tempor invidunt</value>
</field>
<field var="muc#roominfo_description" type="text-single" label="Raum Beschreibung">
<value>Lorem ipsum dolor sit amet, consetetur sadipscing elitr sed diam nonumy eirmod tempor invidunt ut labore et dolore magna</value>
</field>
<field var="muc#roominfo_contactjid" type="jid-multi" label="Contact Addresses (normally, room owner or owners)">
<value>userA@user.us</value>
<value>userB@user.us</value>
</field>
<field var="muc#roominfo_changesubject" type="boolean" label="Occupants May Change the Subject">
<value>1</value>
</field>
<field var="muc#roomconfig_allowinvites" type="boolean" label="Occupants are allowed to invite others">
<value>1</value>
</field>
<field var="muc#roomconfig_allowpm" type="list-single" label="Roles that May Send Private Messages">
<value>anyone</value>
<option label="Anyone">
<value>anyone</value>
</option>
<option label="Anyone with Voice">
<value>participants</value>
</option>
<option label="Moderators Only">
<value>moderators</value>
</option>
<option label="Nobody">
<value>none</value>
</option>
</field>
<field var="muc#roominfo_lang" type="text-single" label="Natural Language for Room Discussions">
<value>de</value>
</field>
<field type="text-single" var="muc#roominfo_logs">
<value>https://logs.xmpp.org/xsf/</value>
</field>
</x>
</query>
</iq>''')

disco_info = parse_disco_info(stanza)

class GroupchatInfo(Gtk.ApplicationWindow):
    def __init__(self):
        Gtk.ApplicationWindow.__init__(self)
        self.set_name('GroupchatJoin')
        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_show_menubar(False)
        self.set_title('Test Group chat info')

        self._main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                                 spacing=18)
        self._main_box.set_valign(Gtk.Align.FILL)

        self._muc_info_box = GroupChatInfoScrolled(None)
        self._muc_info_box.set_vexpand(True)

        self._main_box.add(self._muc_info_box)

        self.add(self._main_box)
        self._muc_info_box.set_from_disco_info(disco_info)
        self._muc_info_box.set_subject(
            'Lorem ipsum dolor sit amet, consetetur sadipscing elitr sed '
            'diam nonumy eirmod tempor invidunt ut labore et dolore magna')
        self._muc_info_box.set_author('userX', None)
        self.show_all()


win = GroupchatInfo()
win.connect('destroy', Gtk.main_quit)
win.show_all()
Gtk.main()



# -*- coding: utf-8 -*-
# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Deletes unneeded DLLs and checks DLL dependencies.
Execute with the build python, will figure out the rest.
"""

import subprocess
import os
import sys

import gi
gi.require_version("GIRepository", "2.0")
from gi.repository import GIRepository


def get_required_by_typelibs():
    deps = set()
    repo = GIRepository.Repository()
    for tl in os.listdir(repo.get_search_path()[0]):
        namespace, version = os.path.splitext(tl)[0].split("-", 1)
        repo.require(namespace, version, 0)
        lib = repo.get_shared_library(namespace)
        if lib:
            deps.update(lib.split(","))
    return deps


EXTENSIONS = [".exe", ".pyd", ".dll"]
SYSTEM_LIBS = ['advapi32.dll',
    "cabinet.dll", "comctl32.dll", "comdlg32.dll", "crypt32.dll", "d3d9.dll",
    "dnsapi.dll", "dsound.dll", "dwmapi.dll", "gdi32.dll", "imm32.dll",
    "iphlpapi.dll", "kernel32.dll", "ksuser.dll", "msi.dll", "msimg32.dll",
    "msvcr71.dll", "msvcr80.dll", "msvcrt.dll", "ole32.dll", "oleaut32.dll",
    "opengl32.dll", "rpcrt4.dll", "setupapi.dll", "shell32.dll", "user32.dll",
    "usp10.dll", "winmm.dll", "winspool.drv", "wldap32.dll", "ws2_32.dll",
    "wsock32.dll", "shlwapi.dll"
]


def get_dependencies(filename):
    deps = []
    try:
        data = subprocess.getoutput("objdump -p %s" % filename)
    except Exception as error:
        print(error)
        return deps

    for line in data.splitlines():
        line = line.strip()
        if line.startswith("DLL Name:"):
            deps.append(line.split(":", 1)[-1].strip().lower())
    return deps


def find_lib(root, name):
    search_path = os.path.join(root, "bin")
    if os.path.exists(os.path.join(search_path, name)):
        return os.path.join(search_path, name)
    elif name in SYSTEM_LIBS:
        return name


def get_things_to_delete(root):
    all_libs = set()
    needed = set()
    for base, dirs, files in os.walk(root):
        for f in files:
            path = os.path.join(base, f)
            if os.path.splitext(path)[-1].lower() in EXTENSIONS:
                all_libs.add(f.lower())
                for lib in get_dependencies(path):
                    all_libs.add(lib)
                    needed.add(lib)
                    if not find_lib(root, lib):
                        print("MISSING:", path, lib)

    for lib in get_required_by_typelibs():
        needed.add(lib)
        if not find_lib(root, lib):
            print("MISSING:", path, lib)

    # get rid of things not in the search path,
    # maybe loaded through other means?
    not_needed = filter(
        lambda l: find_lib(root, l) and \
            os.path.splitext(l)[-1].lower() != ".exe", all_libs - needed)

    return [find_lib(root, l) for l in not_needed]


def main():
    libs = get_things_to_delete(sys.prefix)
    while libs:
        for l in libs:
            print("DELETE:", l)
            os.unlink(l)
        libs = get_things_to_delete(sys.prefix)


if __name__ == "__main__":
    main()



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.

# XEP-0202: Entity Time

import time

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.date_and_time import create_tzinfo

from gajim.common import app
from gajim.common.nec import NetworkEvent
from gajim.common.modules.base import BaseModule


class EntityTime(BaseModule):
    def __init__(self, con):
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._answer_request,
                          typ='get',
                          ns=Namespace.TIME_REVISED),
        ]

    def request_entity_time(self, jid, resource):
        if not app.account_is_available(self._account):
            return

        if resource:
            jid += '/' + resource
        iq = nbxmpp.Iq(to=jid, typ='get')
        iq.addChild('time', namespace=Namespace.TIME_REVISED)

        self._log.info('Requested: %s', jid)

        self._con.connection.SendAndCallForResponse(iq, self._result_received)

    def _result_received(self, _nbxmpp_client, stanza):
        time_info = None
        if not nbxmpp.isResultNode(stanza):
            self._log.info('Error: %s', stanza.getError())
        else:
            time_info = self._extract_info(stanza)

        self._log.info('Received: %s %s', stanza.getFrom(), time_info)

        app.nec.push_incoming_event(NetworkEvent('time-result-received',
                                                 conn=self._con,
                                                 jid=stanza.getFrom(),
                                                 time_info=time_info))

    def _extract_info(self, stanza):
        time_ = stanza.getTag('time')
        if not time_:
            self._log.warning('No time node: %s', stanza)
            return None

        tzo = time_.getTag('tzo').getData()
        if not tzo:
            self._log.warning('Wrong tzo node: %s', stanza)
            return None

        remote_tz = create_tzinfo(tz_string=tzo)
        if remote_tz is None:
            self._log.warning('Wrong tzo node: %s', stanza)
            return None

        utc_time = time_.getTag('utc').getData()
        date_time = parse_datetime(utc_time, check_utc=True)
        if date_time is None:
            self._log.warning('Wrong timezone definition: %s %s',
                              utc_time, stanza.getFrom())
            return None

        date_time = date_time.astimezone(remote_tz)
        return date_time.strftime('%c %Z')

    def _answer_request(self, _con, stanza, _properties):
        self._log.info('%s asked for the time', stanza.getFrom())
        if app.settings.get_account_setting(self._account, 'send_time_info'):
            iq = stanza.buildReply('result')
            time_ = iq.setTag('time', namespace=Namespace.TIME_REVISED)
            formated_time = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
            time_.setTagData('utc', formated_time)
            isdst = time.localtime().tm_isdst
            zone = -(time.timezone, time.altzone)[isdst] / 60.0
            tzo = (zone / 60, abs(zone % 60))
            time_.setTagData('tzo', '%+03d:%02d' % (tzo))
            self._log.info('Answer: %s %s', formated_time, '%+03d:%02d' % (tzo))
        else:
            iq = stanza.buildReply('error')
            err = nbxmpp.ErrorNode(nbxmpp.ERR_SERVICE_UNAVAILABLE)
            iq.addChild(node=err)
            self._log.info('Send service-unavailable')
        self._con.connection.send(iq)
        raise nbxmpp.NodeProcessed


def get_instance(*args, **kwargs):
    return EntityTime(*args, **kwargs), 'EntityTime'



'''
Some diverse tests covering functionality in the GUI Interface class.
'''
import unittest

from test import lib
lib.setup_env()

from gajim.common import logging_helpers
logging_helpers.set_quiet()

from gajim.common import app

from gajim.gui_interface import Interface

from gi.repository import GLib

class TestInterface(unittest.TestCase):

    def test_instantiation(self):
        ''' Test that we can proper initialize and do not fail on globals '''
        def close_app():
            app.app.quit()
        GLib.idle_add(close_app)
        app.app.run()


if __name__ == "__main__":
    unittest.main()



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.

# XEP-0153: vCard-Based Avatars

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.const import AvatarState
from nbxmpp.modules.util import is_error

from gajim.common import app
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import as_task


class VCardAvatars(BaseModule):
    def __init__(self, con):
        BaseModule.__init__(self, con)
        self._requested_shas = []

        self.handlers = [
            StanzaHandler(name='presence',
                          callback=self._presence_received,
                          ns=Namespace.VCARD_UPDATE,
                          priority=51),
        ]

        self.avatar_conversion_available = False

    def pass_disco(self, info):
        is_available = Namespace.VCARD_CONVERSION in info.features
        self.avatar_conversion_available = is_available
        self._log.info('Discovered Avatar Conversion')

    @as_task
    def _request_vcard(self, jid, expected_sha, type_):
        _task = yield

        vcard = yield self._con.get_module('VCardTemp').request_vcard(jid=jid)

        if is_error(vcard):
            self._log.warning(vcard)
            return

        avatar, avatar_sha = vcard.get_avatar()
        if avatar is None:
            self._log.warning('Avatar missing: %s %s', jid, expected_sha)
            return

        if expected_sha != avatar_sha:
            self._log.warning('Avatar mismatch: %s %s != %s',
                              jid,
                              expected_sha,
                              avatar_sha)
            return

        self._log.info('Received: %s %s', jid, avatar_sha)
        app.interface.save_avatar(avatar)

        if type_ == 'contact':
            self._con.get_module('Roster').set_avatar_sha(jid, avatar_sha)
            app.contacts.set_avatar(self._account, jid, avatar_sha)
            app.interface.update_avatar(self._account, jid)

        elif type_ == 'muc':
            app.storage.cache.set_muc_avatar_sha(jid, avatar_sha)
            app.contacts.set_avatar(self._account, jid, avatar_sha)
            app.interface.update_avatar(self._account, jid, room_avatar=True)

        elif type_ == 'muc-user':
            contact = app.contacts.get_gc_contact(self._account,
                                                  jid.bare,
                                                  jid.resource)
            if contact is not None:
                contact.avatar_sha = avatar_sha
                app.interface.update_avatar(contact=contact)

    def _presence_received(self, _con, _stanza, properties):
        if not properties.type.is_available:
            return

        if properties.avatar_state in (AvatarState.IGNORE,
                                       AvatarState.NOT_READY):
            return

        if self._con.get_own_jid().bare_match(properties.jid):
            return

        if properties.from_muc:
            self._gc_update_received(properties)
        else:
            # Check if presence is from a MUC service
            contact = app.contacts.get_groupchat_contact(self._account,
                                                         str(properties.jid))
            self._update_received(properties, room=contact is not None)

    def muc_disco_info_update(self, disco_info):
        if not disco_info.supports(Namespace.VCARD):
            return

        field_var = '{http://modules.prosody.im/mod_vcard_muc}avatar#sha1'
        if not disco_info.has_field(Namespace.MUC_INFO, field_var):
            # Workaround so we don’t delete the avatar for servers that don’t
            # support sha in disco info. Once there is a accepted XEP this
            # can be removed
            return

        avatar_sha = disco_info.get_field_value(Namespace.MUC_INFO, field_var)
        state = AvatarState.EMPTY if not avatar_sha else AvatarState.ADVERTISED
        self._process_update(str(disco_info.jid), state, avatar_sha, True)

    def _update_received(self, properties, room=False):
        self._process_update(properties.jid.bare,
                             properties.avatar_state,
                             properties.avatar_sha,
                             room)

    def _process_update(self, jid, state, avatar_sha, room):
        if state == AvatarState.EMPTY:
            # Empty <photo/> tag, means no avatar is advertised
            self._log.info('%s has no avatar published', jid)
            app.contacts.set_avatar(self._account, jid, None)

            if room:
                app.storage.cache.set_muc_avatar_sha(jid, None)
            else:
                self._con.get_module('Roster').set_avatar_sha(jid, None)
            app.interface.update_avatar(self._account, jid, room_avatar=room)
        else:
            self._log.info('Update: %s %s', jid, avatar_sha)
            current_sha = app.contacts.get_avatar_sha(self._account, jid)

            if avatar_sha == current_sha:
                self._log.info('Avatar already known: %s %s', jid, avatar_sha)
                return

            if app.interface.avatar_exists(avatar_sha):
                # Check if the avatar is already in storage
                self._log.info('Found avatar in storage')
                if room:
                    app.storage.cache.set_muc_avatar_sha(jid, avatar_sha)
                else:
                    self._con.get_module('Roster').set_avatar_sha(jid,
                                                                  avatar_sha)
                app.contacts.set_avatar(self._account, jid, avatar_sha)
                app.interface.update_avatar(
                    self._account, jid, room_avatar=room)
                return

            if avatar_sha not in self._requested_shas:
                self._requested_shas.append(avatar_sha)
                if room:
                    self._request_vcard(jid, avatar_sha, 'muc')
                else:
                    self._request_vcard(jid, avatar_sha, 'contact')

    def _gc_update_received(self, properties):
        nick = properties.jid.resource

        gc_contact = app.contacts.get_gc_contact(
            self._account, properties.jid.bare, nick)

        if gc_contact is None:
            self._log.error('no gc contact found: %s', nick)
            return

        if properties.avatar_state == AvatarState.EMPTY:
            # Empty <photo/> tag, means no avatar is advertised
            self._log.info('%s has no avatar published', nick)
            gc_contact.avatar_sha = None
            app.interface.update_avatar(contact=gc_contact)
        else:
            self._log.info('Update: %s %s', nick, properties.avatar_sha)
            if not app.interface.avatar_exists(properties.avatar_sha):
                if properties.avatar_sha not in self._requested_shas:
                    app.log('avatar').info('Request: %s', nick)
                    self._requested_shas.append(properties.avatar_sha)
                    self._request_vcard(properties.jid,
                                        properties.avatar_sha,
                                        'muc-user')
                return

            if gc_contact.avatar_sha != properties.avatar_sha:
                self._log.info('%s changed their Avatar: %s',
                               nick, properties.avatar_sha)
                gc_contact.avatar_sha = properties.avatar_sha
                app.interface.update_avatar(contact=gc_contact)
            else:
                self._log.info('Avatar already known: %s', nick)


def get_instance(*args, **kwargs):
    return VCardAvatars(*args, **kwargs), 'VCardAvatars'



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from collections import namedtuple

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Pango

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.const import ButtonAction
from gajim.common.helpers import convert_gio_to_openssl_cert

from .util import get_builder
from .util import get_thumbnail_size


class DialogButton(namedtuple('DialogButton', ('response text callback args '
                                               'kwargs action is_default'))):
    @classmethod
    def make(cls, type_=None, **kwargs):
        # Defaults
        default_kwargs = {
            'response': None,
            'text': None,
            'callback': None,
            'args': [],
            'kwargs': {},
            'action': None,
            'is_default': False
        }

        if type_ is not None:
            if type_ == 'OK':
                default_kwargs['response'] = Gtk.ResponseType.OK
                default_kwargs['text'] = _('_OK')

            elif type_ == 'Cancel':
                default_kwargs['response'] = Gtk.ResponseType.CANCEL
                default_kwargs['text'] = _('_Cancel')

            elif type_ == 'Accept':
                default_kwargs['response'] = Gtk.ResponseType.ACCEPT
                default_kwargs['text'] = _('_Accept')
                default_kwargs['is_default'] = True
                default_kwargs['action'] = ButtonAction.SUGGESTED

            elif type_ == 'Delete':
                default_kwargs['response'] = Gtk.ResponseType.REJECT
                default_kwargs['text'] = _('_Delete')
                default_kwargs['action'] = ButtonAction.DESTRUCTIVE

            elif type_ == 'Remove':
                default_kwargs['response'] = Gtk.ResponseType.REJECT
                default_kwargs['text'] = _('_Remove')
                default_kwargs['action'] = ButtonAction.DESTRUCTIVE
            else:
                raise ValueError('Unknown button type: %s ' % type_)

        default_kwargs.update(kwargs)
        return cls(**default_kwargs)


class HigDialog(Gtk.MessageDialog):
    def __init__(self,
                 parent,
                 type_,
                 buttons,
                 pritext,
                 sectext,
                 on_response_ok=None,
                 on_response_cancel=None,
                 on_response_yes=None,
                 on_response_no=None):
        self.call_cancel_on_destroy = True
        Gtk.MessageDialog.__init__(self,
                                   transient_for=parent,
                                   modal=True,
                                   destroy_with_parent=True,
                                   message_type=type_,
                                   buttons=buttons,
                                   text=pritext)

        self.format_secondary_markup(sectext)

        self.possible_responses = {
            Gtk.ResponseType.OK: on_response_ok,
            Gtk.ResponseType.CANCEL: on_response_cancel,
            Gtk.ResponseType.YES: on_response_yes,
            Gtk.ResponseType.NO: on_response_no
        }

        self.connect('response', self.on_response)
        self.connect('destroy', self.on_dialog_destroy)

    def on_response(self, dialog, response_id):
        if response_id not in self.possible_responses:
            return
        if not self.possible_responses[response_id]:
            self.destroy()
        elif isinstance(self.possible_responses[response_id], tuple):
            if len(self.possible_responses[response_id]) == 1:
                self.possible_responses[response_id][0](dialog)
            else:
                self.possible_responses[response_id][0](
                    dialog, *self.possible_responses[response_id][1:])
        else:
            self.possible_responses[response_id](dialog)

    def on_dialog_destroy(self, _widget):
        if not self.call_cancel_on_destroy:
            return None
        cancel_handler = self.possible_responses[Gtk.ResponseType.CANCEL]
        if not cancel_handler:
            return False
        if isinstance(cancel_handler, tuple):
            cancel_handler[0](None, *cancel_handler[1:])
        else:
            cancel_handler(None)
        return None

    def popup(self):
        """
        Show dialog
        """
        vb = self.get_children()[0].get_children()[0] # Give focus to top vbox
        # vb.set_flags(Gtk.CAN_FOCUS)
        vb.grab_focus()
        self.show_all()


class WarningDialog(HigDialog):
    """
    HIG compliant warning dialog
    """

    def __init__(self, pritext, sectext='', transient_for=None):
        if transient_for is None:
            transient_for = app.app.get_active_window()
        HigDialog.__init__(self,
                           transient_for,
                           Gtk.MessageType.WARNING,
                           Gtk.ButtonsType.OK,
                           pritext,
                           sectext)
        self.set_modal(False)
        self.popup()


class InformationDialog(HigDialog):
    """
    HIG compliant info dialog
    """

    def __init__(self, pritext, sectext='', transient_for=None):
        if transient_for is None:
            transient_for = app.app.get_active_window()
        HigDialog.__init__(self,
                           transient_for,
                           Gtk.MessageType.INFO,
                           Gtk.ButtonsType.OK,
                           pritext,
                           sectext)
        self.set_modal(False)
        self.popup()


class ErrorDialog(HigDialog):
    """
    HIG compliant error dialog
    """

    def __init__(self,
                 pritext,
                 sectext='',
                 on_response_ok=None,
                 on_response_cancel=None,
                 transient_for=None):
        if transient_for is None:
            transient_for = app.app.get_active_window()
        HigDialog.__init__(self,
                           transient_for,
                           Gtk.MessageType.ERROR,
                           Gtk.ButtonsType.OK,
                           pritext,
                           sectext,
                           on_response_ok=on_response_ok,
                           on_response_cancel=on_response_cancel)
        self.popup()


class CertificateDialog(Gtk.ApplicationWindow):
    def __init__(self, transient_for, account, cert):
        Gtk.ApplicationWindow.__init__(self)
        self.account = account
        self.set_name('CertificateDialog')
        self.set_application(app.app)
        self.set_show_menubar(False)
        self.set_resizable(False)
        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_title(_('Certificate'))

        self._ui = get_builder('certificate_dialog.ui')
        self.add(self._ui.certificate_box)

        self.connect('key-press-event', self._on_key_press)

        self._clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        cert = convert_gio_to_openssl_cert(cert)
        # Get data for labels and copy button
        issuer = cert.get_issuer()
        subject = cert.get_subject()

        self._headline = _('Certificate for \n%s') % self.account
        self._it_common_name = subject.commonName or ''
        self._it_organization = subject.organizationName or ''
        self._it_org_unit = subject.organizationalUnitName or ''
        it_serial_no = str(cert.get_serial_number())
        it_serial_no_half = int(len(it_serial_no) / 2)
        self._it_serial_number = '%s\n%s' % (
            it_serial_no[:it_serial_no_half],
            it_serial_no[it_serial_no_half:])
        self._ib_common_name = issuer.commonName or ''
        self._ib_organization = issuer.organizationName or ''
        self._ib_org_unit = issuer.organizationalUnitName or ''
        issued = datetime.strptime(cert.get_notBefore().decode('ascii'),
                                   '%Y%m%d%H%M%SZ')
        self._issued = issued.strftime('%c %Z')
        expires = datetime.strptime(cert.get_notAfter().decode('ascii'),
                                    '%Y%m%d%H%M%SZ')
        self._expires = expires.strftime('%c %Z')
        sha1 = cert.digest('sha1').decode('utf-8')
        self._sha1 = '%s\n%s' % (sha1[:29], sha1[30:])
        sha256 = cert.digest('sha256').decode('utf-8')
        self._sha256 = '%s\n%s\n%s\n%s' % (
            sha256[:23], sha256[24:47], sha256[48:71], sha256[72:])

        # Set labels
        self._ui.label_cert_for_account.set_text(self._headline)
        self._ui.data_it_common_name.set_text(self._it_common_name)
        self._ui.data_it_organization.set_text(self._it_organization)
        self._ui.data_it_organizational_unit.set_text(self._it_org_unit)
        self._ui.data_it_serial_number.set_text(self._it_serial_number)
        self._ui.data_ib_common_name.set_text(self._ib_common_name)
        self._ui.data_ib_organization.set_text(self._ib_organization)
        self._ui.data_ib_organizational_unit.set_text(self._ib_org_unit)
        self._ui.data_issued_on.set_text(self._issued)
        self._ui.data_expires_on.set_text(self._expires)
        self._ui.data_sha1.set_text(self._sha1)
        self._ui.data_sha256.set_text(self._sha256)

        self.set_transient_for(transient_for)
        self._ui.connect_signals(self)
        self.show_all()

    def _on_key_press(self, _widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.destroy()

    def _on_copy_cert_info_button_clicked(self, _widget):
        clipboard_text = \
            self._headline + '\n\n' + \
            _('Issued to\n') + \
            _('Common Name (CN): ') + self._it_common_name + '\n' + \
            _('Organization (O): ') + self._it_organization + '\n' + \
            _('Organizational Unit (OU): ') + self._it_org_unit + '\n' + \
            _('Serial Number: ') + self._it_serial_number + '\n\n' + \
            _('Issued by\n') + \
            _('Common Name (CN): ') + self._ib_common_name + '\n' + \
            _('Organization (O): ') + self._ib_organization + '\n' + \
            _('Organizational Unit (OU): ') + self._ib_org_unit + '\n\n' + \
            _('Validity\n') + \
            _('Issued on: ') + self._issued + '\n' + \
            _('Expires on: ') + self._expires + '\n\n' + \
            _('SHA-1:') + '\n' + \
            self._sha1 + '\n' + \
            _('SHA-256:') + '\n' + \
            self._sha256 + '\n'
        self._clipboard.set_text(clipboard_text, -1)


class PassphraseDialog:
    """
    Class for Passphrase dialog
    """
    def __init__(self, titletext, labeltext, checkbuttontext=None,
                 ok_handler=None, cancel_handler=None, transient_for=None):
        self._ui = get_builder('passphrase_dialog.ui')
        self.window = self._ui.get_object('passphrase_dialog')
        self.passphrase = -1
        self.window.set_title(titletext)
        self._ui.message_label.set_text(labeltext)

        self._ok = False

        self.cancel_handler = cancel_handler
        self.ok_handler = ok_handler
        self._ui.ok_button.connect('clicked', self.on_okbutton_clicked)
        self._ui.cancel_button.connect('clicked', self.on_cancelbutton_clicked)

        self._ui.connect_signals(self)
        if transient_for is None:
            transient_for = app.app.get_active_window()
        self.window.set_transient_for(transient_for)
        self.window.show_all()

        self.check = bool(checkbuttontext)
        if self._ui.save_passphrase_checkbutton:
            self._ui.save_passphrase_checkbutton.set_label(checkbuttontext)
        else:
            self._ui.save_passphrase_checkbutton.hide()

    def on_okbutton_clicked(self, _widget):
        if not self.ok_handler:
            return

        passph = self._ui.passphrase_entry.get_text()

        if self.check:
            checked = self._ui.save_passphrase_checkbutton.get_active()
        else:
            checked = False

        self._ok = True

        self.window.destroy()

        if isinstance(self.ok_handler, tuple):
            self.ok_handler[0](passph, checked, *self.ok_handler[1:])
        else:
            self.ok_handler(passph, checked)

    def on_cancelbutton_clicked(self, _widget):
        self.window.destroy()

    def on_passphrase_dialog_destroy(self, _widget):
        if self.cancel_handler and not self._ok:
            self.cancel_handler()


class ConfirmationDialog(Gtk.MessageDialog):
    def __init__(self, title, text, sec_text, buttons,
                 modal=True, transient_for=None):
        if transient_for is None:
            transient_for = app.app.get_active_window()
        Gtk.MessageDialog.__init__(self,
                                   title=title,
                                   text=text,
                                   transient_for=transient_for,
                                   message_type=Gtk.MessageType.QUESTION,
                                   modal=modal)

        self.get_style_context().add_class('confirmation-dialog')

        self._buttons = {}

        for button in buttons:
            self._buttons[button.response] = button
            self.add_button(button.text, button.response)
            if button.is_default:
                self.set_default_response(button.response)
            if button.action is not None:
                widget = self.get_widget_for_response(button.response)
                widget.get_style_context().add_class(button.action.value)

        self.format_secondary_markup(sec_text)

        self.connect('response', self._on_response)

    def _on_response(self, _dialog, response):
        if response == Gtk.ResponseType.DELETE_EVENT:
            # Look if DELETE_EVENT is mapped to another response
            response = self._buttons.get(response, None)
            if response is None:
                # If DELETE_EVENT was not mapped we assume CANCEL
                response = Gtk.ResponseType.CANCEL

        button = self._buttons.get(response, None)
        if button is None:
            self.destroy()
            return

        if button.callback is not None:
            button.callback(*button.args, **button.kwargs)
        self.destroy()

    def show(self):
        self.show_all()


NewConfirmationDialog = ConfirmationDialog


class ConfirmationCheckDialog(ConfirmationDialog):
    def __init__(self, title, text, sec_text, check_text,
                 buttons, modal=True, transient_for=None):
        ConfirmationDialog.__init__(self,
                                    title,
                                    text,
                                    sec_text,
                                    buttons,
                                    transient_for=transient_for,
                                    modal=modal)

        self._checkbutton = Gtk.CheckButton.new_with_mnemonic(check_text)
        self._checkbutton.set_can_focus(False)
        self._checkbutton.set_margin_start(30)
        self._checkbutton.set_margin_end(30)
        label = self._checkbutton.get_child()
        label.set_line_wrap(True)
        label.set_max_width_chars(50)
        label.set_halign(Gtk.Align.START)
        label.set_line_wrap_mode(Pango.WrapMode.WORD)
        label.set_margin_start(10)

        self.get_content_area().add(self._checkbutton)

    def _on_response(self, _dialog, response):
        button = self._buttons.get(response)
        if button is not None:
            button.args.insert(0, self._checkbutton.get_active())
        super()._on_response(_dialog, response)


NewConfirmationCheckDialog = ConfirmationCheckDialog


class PastePreviewDialog(ConfirmationCheckDialog):
    def __init__(self, title, text, sec_text, check_text, image,
                 buttons, modal=True, transient_for=None):
        ConfirmationCheckDialog.__init__(self,
                                         title,
                                         text,
                                         sec_text,
                                         check_text,
                                         buttons,
                                         transient_for=transient_for,
                                         modal=modal)

        preview = Gtk.Image()
        preview.set_halign(Gtk.Align.CENTER)
        preview.get_style_context().add_class('preview-image')
        size = 300
        image_width = image.get_width()
        image_height = image.get_height()

        if size > image_width and size > image_height:
            preview.set_from_pixbuf(image)
        else:
            thumb_width, thumb_height = get_thumbnail_size(image, size)
            pixbuf_scaled = image.scale_simple(
                thumb_width, thumb_height, GdkPixbuf.InterpType.BILINEAR)
            preview.set_from_pixbuf(pixbuf_scaled)

        content_area = self.get_content_area()
        content_area.pack_start(preview, True, True, 0)
        content_area.reorder_child(preview, 2)


class InputDialog(ConfirmationDialog):
    def __init__(self, title, text, sec_text, buttons, input_str=None,
                 transient_for=None, modal=True):
        ConfirmationDialog.__init__(self,
                                    title,
                                    text,
                                    sec_text,
                                    buttons,
                                    transient_for=transient_for,
                                    modal=modal)

        self._entry = Gtk.Entry()
        self._entry.set_activates_default(True)
        self._entry.set_margin_start(50)
        self._entry.set_margin_end(50)

        if input_str:
            self._entry.set_text(input_str)
            self._entry.select_region(0, -1)  # select all

        self.get_content_area().add(self._entry)

    def _on_response(self, _dialog, response):
        button = self._buttons.get(response)
        if button is not None:
            button.args.insert(0, self._entry.get_text())
        super()._on_response(_dialog, response)


class TimeoutWindow:
    """
    Class designed to be derivated by other windows
    Derived windows close automatically after reaching the timeout
    """
    def __init__(self, timeout):
        self.title_text = ''
        self._countdown_left = timeout
        self._timeout_source_id = None

    def start_timeout(self):
        if self._countdown_left > 0:
            self.countdown()
            self._timeout_source_id = GLib.timeout_add_seconds(
                1, self.countdown)

    def stop_timeout(self, *args, **kwargs):
        if self._timeout_source_id is not None:
            GLib.source_remove(self._timeout_source_id)
            self._timeout_source_id = None
        self.set_title(self.title_text)

    def on_timeout(self):
        """
        To be implemented by derivated classes
        """

    def countdown(self):
        if self._countdown_left <= 0:
            self._timeout_source_id = None
            self.on_timeout()
            return False

        self.set_title('%s [%s]' % (
            self.title_text, str(self._countdown_left)))
        self._countdown_left -= 1
        return True


class ShortcutsWindow:
    def __init__(self):
        transient = app.app.get_active_window()
        builder = get_builder('shortcuts_window.ui')
        self.window = builder.get_object('shortcuts_window')
        self.window.connect('destroy', self._on_window_destroy)
        self.window.set_transient_for(transient)
        self.window.show_all()
        self.window.present()

    def _on_window_destroy(self, _widget):
        self.window = None



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.

# XEP-0084: User Avatar

from nbxmpp.namespaces import Namespace
from nbxmpp.modules.util import is_error

from gajim.common import app
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node
from gajim.common.modules.util import as_task

class UserAvatar(BaseModule):

    _nbxmpp_extends = 'UserAvatar'
    _nbxmpp_methods = [
        'request_avatar_metadata',
        'request_avatar_data',
        'set_avatar',
        'set_access_model'
    ]

    def __init__(self, con):
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._avatar_metadata_received)

    @event_node(Namespace.AVATAR_METADATA)
    def _avatar_metadata_received(self, _con, _stanza, properties):
        if properties.pubsub_event.retracted:
            return

        metadata = properties.pubsub_event.data
        jid = str(properties.jid)

        if metadata is None or not metadata.infos:
            self._log.info('No avatar published: %s', jid)
            app.contacts.set_avatar(self._account, jid, None)
            self._con.get_module('Roster').set_avatar_sha(jid, None)
            app.interface.update_avatar(self._account, jid)
        else:
            if properties.is_self_message:
                sha = app.settings.get_account_setting(self._account,
                                                       'avatar_sha')
            else:
                sha = app.contacts.get_avatar_sha(self._account, jid)

            if sha in metadata.avatar_shas:
                self._log.info('Avatar already known: %s %s', jid, sha)
                return

            avatar_info = metadata.infos[0]
            self._log.info('Request: %s %s', jid, avatar_info.id)
            self._request_avatar_data(jid, avatar_info)

    @as_task
    def _request_avatar_data(self, jid, avatar_info):
        _task = yield

        avatar = yield self._nbxmpp('UserAvatar').request_avatar_data(
            avatar_info.id, jid=jid)

        if avatar is None:
            self._log.warning('%s advertised %s but data node is empty',
                              jid, avatar_info.id)
            return

        if is_error(avatar):
            self._log.warning(avatar)
            return

        self._log.info('Received Avatar: %s %s', jid, avatar.sha)
        app.interface.save_avatar(avatar.data)

        if self._con.get_own_jid().bare_match(jid):
            app.settings.set_account_setting(self._account,
                                             'avatar_sha',
                                             avatar.sha)
        else:
            self._con.get_module('Roster').set_avatar_sha(
                str(jid), avatar.sha)

        app.contacts.set_avatar(self._account, str(jid), avatar.sha)
        app.interface.update_avatar(self._account, str(jid))


def get_instance(*args, **kwargs):
    return UserAvatar(*args, **kwargs), 'UserAvatar'



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.


import logging

from gi.repository import GLib
from gi.repository import Gio

from gajim.common import configpaths

from gajim.common.helpers import get_random_string
from gajim.common.helpers import write_file_async


log = logging.getLogger('gajim.c.cert_store')


class CertificateStore:
    def __init__(self):
        self._path = configpaths.get('CERT_STORE')
        self._certs = []

        self._load_certificates()

    def _get_random_path(self):
        filename = get_random_string()
        path = self._path / filename
        if path.exists():
            return self._get_random_path()
        return path

    def _load_certificates(self):
        for path in self._path.iterdir():
            if path.is_dir():
                continue
            try:
                cert = Gio.TlsCertificate.new_from_file(str(path))
            except GLib.Error as error:
                log.warning('Can\'t load certificate: %s, %s', path, error)
                continue

            log.info('Loaded: %s', path.stem)
            self._certs.append(cert)

        log.info('%s Certificates loaded', len(self._certs))

    def get_certificates(self):
        return list(self._certs)

    def add_certificate(self, certificate):
        log.info('Add certificate to trust store')
        self._certs.append(certificate)
        pem = certificate.props.certificate_pem
        path = self._get_random_path()
        write_file_async(path,
                         pem.encode(),
                         self._on_certificate_write_finished,
                         path)

    def verify(self, certificate, tls_errors):
        if Gio.TlsCertificateFlags.UNKNOWN_CA in tls_errors:
            for trusted_certificate in self._certs:
                if trusted_certificate.is_same(certificate):
                    tls_errors.remove(Gio.TlsCertificateFlags.UNKNOWN_CA)
                    break

        if not tls_errors:
            return True
        return False

    @staticmethod
    def _on_certificate_write_finished(data, error, path):
        if data is None:
            log.error('Can\'t store certificate: %s', error)
            return

        log.info('Certificate stored: %s', path)



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.

# XEP-0118: User Tune

from typing import Any
from typing import Tuple

from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common import ged
from gajim.common.nec import NetworkEvent
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node
from gajim.common.modules.util import store_publish
from gajim.common.const import PEPEventType
from gajim.common.dbus.music_track import MusicTrackListener
from gajim.common.helpers import event_filter


class UserTune(BaseModule):

    _nbxmpp_extends = 'Tune'
    _nbxmpp_methods = [
        'set_tune',
    ]

    def __init__(self, con):
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._tune_received)
        self._tune_data = None

        self.register_events([
            ('music-track-changed', ged.CORE, self._on_music_track_changed),
            ('signed-in', ged.CORE, self._on_signed_in),
        ])

    def get_current_tune(self):
        return self._tune_data

    @event_node(Namespace.TUNE)
    def _tune_received(self, _con, _stanza, properties):
        if properties.pubsub_event.retracted:
            return

        data = properties.pubsub_event.data
        for contact in app.contacts.get_contacts(self._account,
                                                 str(properties.jid)):
            if data is not None:
                contact.pep[PEPEventType.TUNE] = data
            else:
                contact.pep.pop(PEPEventType.TUNE, None)

        if properties.is_self_message:
            if data is not None:
                self._con.pep[PEPEventType.TUNE] = data
            else:
                self._con.pep.pop(PEPEventType.TUNE, None)
            self._tune_data = data

        app.nec.push_incoming_event(
            NetworkEvent('tune-received',
                         account=self._account,
                         jid=properties.jid.bare,
                         tune=data,
                         is_self_message=properties.is_self_message))

    @store_publish
    def set_tune(self, tune):
        if not self._con.get_module('PEP').supported:
            return

        if not app.settings.get_account_setting(self._account, 'publish_tune'):
            return

        if tune == self._tune_data:
            return

        self._tune_data = tune

        self._log.info('Send %s', tune)
        self._nbxmpp('Tune').set_tune(tune)

    def set_enabled(self, enable):
        if enable:
            app.settings.set_account_setting(self._account,
                                             'publish_tune',
                                             True)
            self._publish_current_tune()

        else:
            self.set_tune(None)
            app.settings.set_account_setting(self._account,
                                             'publish_tune',
                                             False)

    def _publish_current_tune(self):
        self.set_tune(MusicTrackListener.get().current_tune)

    @event_filter(['account'])
    def _on_signed_in(self, _event):
        self._publish_current_tune()

    def _on_music_track_changed(self, event):
        if self._tune_data == event.info:
            return

        self.set_tune(event.info)


def get_instance(*args: Any, **kwargs: Any) -> Tuple[UserTune, str]:
    return UserTune(*args, **kwargs), 'UserTune'
